from typing import List, Optional, Union, Dict, Callable, Tuple
from itertools import repeat
import os
import sys
import copy
import json
import time
import uuid

import numpy as np

import ROOT

import quickstats
from quickstats import semistaticmethod, Timer
from quickstats.components import ROOTObject
from quickstats.components.modelling import TreeDataSource, PdfFitTool
from quickstats.utils.common_utils import combine_dict, execute_multi_tasks

class DataModelling(ROOTObject):
    
    _DEFAULT_FIT_OPTION_ = {
        'print_level': -1,
        'min_fit': 2,
        'max_fit': 3,
        'binned': False,
        'minos': False,
        'hesse': True,
        'sumw2': True,
        'strategy': 1,
        'range_expand_rate': 1        
    }
    
    _DEFAULT_PLOT_OPTION_ = {
        'bin_range': None,
        'n_bins_data': None,
        'n_bins_pdf': 1000,
        'bin_error': True,
        'show_comparison': True,
        'show_params': True,
        'show_stats': True,
        'show_error': True,        
        'value_fmt': "{:.2f}",
        'stats_list': ["chi2/ndf"],
        'comparison_options':{
            "styles": {
              "color": "k"  
            },            
            "mode": "difference",
            "label": "Fit - MC"            
        },
        'draw_options':{},
        'styles': None,
        'xlabel': None,
        'ylabel': "Events / ({bin_width:.2f})",
        'annotations': None,
        'label_map': {
            'data' : "MC",
            'pdf'  : "Fit"
        }
    }

    # pdf class defined in macros
    _EXTERNAL_PDF_ = ['RooTwoSidedCBShape']
    
    _PDF_MAP_ = {
        'RooCrystalBall_DSCB' : 'RooCrystalBall',
        'DSCB'                : 'RooTwoSidedCBShape',
        'ExpGaussExp'         : 'RooExpGaussExpShape',
        'Exp'                 : 'RooExponential',
        'Bukin'               : 'RooBukinPdf',
        'Gaussian'            : 'RooGaussian',
        'Gaus'                : 'RooGaussian'
    }
    
    _DEFAULT_ROOT_CONFIG_ = {
        "SetBatch" : True,
        "TH1Sumw2" : True
    }
    
    _REQUIRE_CONFIG_ = {
        "ROOT"  : True,
        "RooFit": True
    }
    
    @property
    def plot_options(self):
        return self._plot_options
    
    @property
    def fit_options(self):
        return self._fit_options
    
    @property
    def model_class(self):
        return self._model_class
    
    @property
    def param_templates(self):
        return self._param_templates

    def __init__(self, observable:str, fit_range:Union[List[float], Tuple[float]],
                 functional_form:Union[str, Callable],
                 param_templates:Optional[Callable]=None,
                 weight:str="weight", n_bins:Optional[int]=None,
                 fit_options:Optional[Dict]=None,
                 plot_options:Optional[Dict]=None,
                 verbosity:str="INFO"):
        """
        Modelling of a data distribution by a simple analytic function.
        
        Parameters:
            observable: str
                Name of observable.
        """
        self._fit_options  = self._DEFAULT_FIT_OPTION_
        self._plot_options = self._DEFAULT_PLOT_OPTION_
        _fit_options = {
            'functional_form' : functional_form,
            'observable'      : observable,
            'range'           : fit_range,
            'n_bins'          : n_bins,
            'weight'          : weight
        }
        fit_options = combine_dict(_fit_options, fit_options)
        self.update_fit_options(fit_options)
        self.update_plot_options(plot_options)
        self.set_param_templates(param_templates)
        self.set_functional_form(functional_form)
        roofit_config = {
            "MinimizerPrintLevel": self.fit_options.get("print_level", -1)
        }
        super().__init__(roofit_config=roofit_config,
                         verbosity=verbosity)
        
    def update_fit_options(self, options:Optional[Dict]=None):
        self._fit_options = combine_dict(self._fit_options, options)
        
    def update_plot_options(self, options:Optional[Dict]=None):
        self._plot_options = combine_dict(self._plot_options, options)        
        
    def set_param_templates(self, param_templates:Callable):
        self._param_templates = param_templates
        
    def set_functional_form(self, functional_form:Union[str, Callable]):
        self._model_class = self.get_model_class(functional_form)
        if self.param_templates is None:
            if isinstance(functional_form, str):
                param_templates = self.get_param_templates(functional_form)
                self.set_param_templates(param_templates)
            else:
                raise RuntimeError("missing parameter templates definition")
                
    @semistaticmethod
    def get_model_class(self, name:Union[str, Callable]):
        if isinstance(name, Callable):
            return name
        pdf_name = self._PDF_MAP_.get(name, name)
        if not hasattr(ROOT, pdf_name):
            if pdf_name in self._EXTERNAL_PDF_:
                self.load_extension(pdf_name)
            else:
                raise ValueError(f"{name} is not a valid pdf name defined by ROOT")
        if not hasattr(ROOT, pdf_name):
            raise RuntimeError(f"failed to load pdf {pdf_name}")
        pdf = getattr(ROOT, pdf_name)
        return pdf
    
    @semistaticmethod
    def get_param_templates(self, name:Union[str, Callable]):
        if isinstance(name, Callable):
            return name
        from .parameter_templates import get_param_templates
        return get_param_templates(name)

    def sanity_check(self):
        if self.model_class is None:
            raise RuntimeError("model pdf not set")
        if self.param_templates is None:
            raise RuntimeError("model parameter templates not set")
    
    @staticmethod
    def get_param_data(parameters:List["ROOT.RooRealVar"], value_only:bool=False):
        param_data = {}
        for name in parameters:
            if value_only:
                param_data[name] = parameters[name].getVal()
            else:
                param_data[name] = {
                    'value'  : parameters[name].getVal(),
                    'errorhi': parameters[name].getErrorHi(),
                    'errorlo': parameters[name].getErrorLo(),
                    'error'  : parameters[name].getError()
                }
        return param_data

    @semistaticmethod
    def _run(self, filename:str, tree_name:str, model_class:Callable, param_templates:Callable,
             fit_options:Dict, plot_options:Optional[Dict]=None, save_summary_as:Optional[str]=None, 
             save_param_as:Optional[str]=None, save_plot_as:Optional[str]=None):
        
        with Timer() as t:
            canvas = ROOT.TCanvas(uuid.uuid4().hex)
            observable_config = {
                'name'        : fit_options['observable'],
                'range'       : fit_options['range'],
                'n_bins'      : fit_options['n_bins']
            }
            data_source = TreeDataSource(tree_name, filename,
                                         observable=observable_config,
                                         weight_name=fit_options['weight'])
            # create dataset
            dataset_name = f"dataset_{observable_config['name']}"
            dataset = data_source.create_dataset(dataset_name)
            observable = dataset.get().first()
            
            # create model pdf
            model_name = f"model_{model_class.Class_Name()}"
            model_parameters = param_templates(data_source)
            model_pdf = model_class(model_name, model_name, observable, *model_parameters.values())

            fit_kwargs = {
                'minos'             : fit_options['minos'],
                'hesse'             : fit_options['hesse'],
                'strategy'          : fit_options['strategy'],
                'sumw2'             : fit_options['sumw2'],
                'min_fit'           : fit_options['min_fit'],
                'max_fit'           : fit_options['max_fit'],
                'range_expand_rate' : fit_options['range_expand_rate'],
                'print_level'       : fit_options['print_level']
            }

            fit_tool   = PdfFitTool(model_pdf, dataset, verbosity=self.stdout.verbosity)
            fit_result = fit_tool.mle_fit(**fit_kwargs)
            fit_result.Print()

            n_float_params = fit_result.floatParsFinal().getSize()
            fit_stats = fit_tool.get_fit_stats(n_float_params=n_float_params)
            fit_tool.print_fit_stats(fit_stats)

            # free memory
            canvas.Close()
            ROOT.gSystem.ProcessEvents()
            
        self.stdout.info(f"INFO: Task finished. Total time taken: {t.interval:.3f}s")
        
        param_data = self.get_param_data(model_parameters)
        param_data_value_only = self.get_param_data(model_parameters, value_only=True)
        summary = {"filename"    : filename,
                   "parameters"  : param_data,
                   "fit_options" : copy.deepcopy(fit_options),
                   "stats"       : fit_stats,
                   "time"        : t.interval}
        if not isinstance(summary['fit_options']['functional_form'], str):
            summary['fit_options']['functional_form'] = type(summary['fit_options']['functional_form']).__name__
            
        if save_plot_as is not None:
            self.create_plot(dataset, model_pdf, observable,
                             plot_options=plot_options,
                             model_summary=summary,
                             save_as=save_plot_as)

        if save_param_as is not None:
            with open(save_param_as, "w") as outfile:
                json.dump(param_data_value_only, outfile, indent=2)
                
        if save_summary_as is not None:
            with open(save_summary_as, "w") as outfile:
                json.dump(summary, outfile, indent=2)

        return summary
    
    @semistaticmethod
    def create_plot(self, data:ROOT.RooDataSet, pdf:ROOT.RooAbsPdf,
                    observable:Optional[ROOT.RooRealVar]=None,
                    plot_options:Optional[Dict]=None,
                    model_summary:Optional[Dict]=None,
                    save_as:Optional[str]=None):
        
        if observable is None:
            parameters = data.get()
            if len(parameters) > 1:
                raise RuntimeError("only single-observable dataset is allowed")
            observable = parameters.first()
        
        plot_options = combine_dict(self._DEFAULT_PLOT_OPTION_, plot_options)
            
        init_kwargs = {
            "bin_range"   : plot_options["bin_range"],
            "n_bins_data" : plot_options["n_bins_data"],
            "n_bins_pdf"  : plot_options["n_bins_pdf"],
            "bin_error"   : plot_options["bin_error"],
            "label_map"   : plot_options["label_map"],
            "color_cycle" : plot_options.get("color_cycle", None),
            "styles"      : plot_options.get("styles", None),
            "analysis_label_options" : plot_options.get("analysis_label_options", None),
            "config"      : plot_options.get("config", None)
        }
        
        from quickstats.plots import PdfDistributionPlot
        plotter = PdfDistributionPlot.from_roofit_data(data, pdf, observable, **init_kwargs)
        
        if model_summary is not None:
            value_fmt = plot_options["value_fmt"]
            annotation_text = ""
            if plot_options['show_params']:
                param_data = model_summary["parameters"]
                for param in param_data:
                    value = value_fmt.format(param_data[param]["value"])
                    if plot_options["show_error"]:
                        error = value_fmt.format(param_data[param]["error"])
                        annotation_text += f"{param} = {value} $\\pm$ {error}\n"
                    else:
                        annotation_text += f"{param} = {value}\n"
                annotation_text += "\n"
            if plot_options["show_stats"]:
                stats_list = plot_options["stats_list"]
                for stat in stats_list:
                    if stat not in model_summary["stats"]:
                        raise RuntimeError(f"invalid stats item: {stat}")
                    value = value_fmt.format(model_summary["stats"][stat])
                    annotation_text += f"{stat} = {value}\n"
                annotation_text += "\n"
            plotter.add_annotation(annotation_text)
        
        annotations = plot_options.get("annotations", None)
        if annotations is not None:
            if isinstance(annotations, dict):
                plotter.add_annotation(**annotations)
            elif isinstance(annotations, list):
                for annotation in annotations:
                    plotter.add_annotation(**annotation)
            else:
                raise ValueError('invalid format for the fit option "annotations"')
        
        xlabel, ylabel = plot_options["xlabel"], plot_options["ylabel"]
        if (xlabel is None):
            xlabel = observable.GetName()
        # deduce bin width from data
        if (ylabel is not None) and "bin_width" in ylabel:
            bin_width = np.unique(np.diff(plotter.collective_data['data']['x']))
            if (len(bin_width) > 1) and (not np.allclose(bin_width, bin_width[0])):
                raise RuntimeError("can not deduce bin width: non-uniform binnings detected")
            bin_width = bin_width[0]
            ylabel = ylabel.format(bin_width=bin_width)            
        
        if plot_options["show_comparison"]:
            comparison_options = {
                "reference": "data",
                "target": "pdf_data_binning"
            }
            comparison_options = combine_dict(comparison_options, plot_options["comparison_options"])
        else:
            comparison_options = None
            
        draw_options = plot_options.get("draw_options", {})

        ax = plotter.draw(xlabel=xlabel, ylabel=ylabel, targets=["data", "pdf"],
                          comparison_options=comparison_options, **draw_options)
        
        import matplotlib.pyplot as plt
        if save_as is not None:
            plt.savefig(save_as, bbox_inches="tight")
        plt.show()

    def run(self, filename:str, tree_name:str, save_summary_as:Optional[str]=None, 
            save_param_as:Optional[str]=None, save_plot_as:Optional[str]=None):
        summary = self._run(filename, tree_name, model_class=self.model_class, 
                            param_templates=self.param_templates,
                            fit_options=self.fit_options,
                            plot_options=self.plot_options,
                            save_summary_as=save_summary_as,
                            save_param_as=save_param_as,
                            save_plot_as=save_plot_as)
        return summary
    
    @staticmethod
    def _get_iter_save_paths(filenames:List[str], save_paths:Optional[Union[List[str], str]]=None):
        # do not save
        if save_paths is None:
            return repeat(None)
        # save by expression
        elif isinstance(save_paths, str):
            _save_paths = []
            for filename in filenames:
                basename = os.path.splitext(os.path.basename(filename))[0]
                _save_paths.append(save_paths.format(basename=basename))
            return _save_paths
        # save by explicit paths
        elif isinstance(save_paths, list):
            if len(filenames) != len(save_paths):
                raise RuntimeError("number of output file paths must match the number input files")
            return save_paths
        else:
            raise ValueError("invalid save option")
        
    def batch_run(self, filenames:List[str], tree_name:str,
                  save_summary_as:Optional[Union[List[str], str]]=None,
                  save_param_as:Optional[Union[List[str], str]]=None,
                  save_plot_as:Optional[Union[List[str], str]]=None,
                  plot_options:Optional[List]=None, parallel:int=-1):
        self.sanity_check()
        save_summary_as = self._get_iter_save_paths(filenames, save_summary_as)
        save_param_as   = self._get_iter_save_paths(filenames, save_param_as)
        save_plot_as    = self._get_iter_save_paths(filenames, save_plot_as)
        if plot_options is None:
            plot_options = repeat(self.plot_options)
        elif isinstance(plot_options, dict):
            plot_options = repeat(combine_dict(self.plot_options, plot_options))
        elif isinstance(plot_options, list):
            plot_options = [combine_dict(self.plot_options, _plot_options) for _plot_options in plot_options]
        else:
            raise ValueError("invalid plot options")
        args = (filenames, repeat(tree_name), repeat(self.model_class),
                repeat(self.param_templates), repeat(self.fit_options),
                plot_options, save_summary_as, save_param_as, 
                save_plot_as)
        results = execute_multi_tasks(self._run, *args, parallel=parallel)
        return results
    
    def run_over_categories(self, prefix:str, categories:List[str],
                            tree_name:str, input_dir:str="./",
                            save_summary_as:Optional[Union[List[str], str]]=None,
                            save_param_as:Optional[Union[List[str], str]]=None,
                            save_plot_as:Optional[Union[List[str], str]]=None,
                            save_merged_param_as:str="model_parameters.json",
                            plot_options:Optional[List]=None,
                            parallel:int=-1):
        filenames = []
        for category in categories:
            filename = os.path.join(input_dir, f"{prefix}_{category}.root")
            filenames.append(filename)

        results = self.batch_run(filenames, tree_name,
                                 save_summary_as=save_summary_as,
                                 save_param_as=save_param_as,
                                 save_plot_as=save_plot_as,
                                 plot_options=plot_options,
                                 parallel=parallel)
        
        # merge parameter data in each category
        category_results = dict(zip(categories, results))
        if save_merged_param_as is not None:
            param_data = {}
            for category, result in category_results.items():
                param_data[category] = {name: data["value"] for name, data in result['parameters'].items()}
            with open(save_merged_param_as, "w") as outfile:
                json.dump(param_data, outfile, indent=2)
        return category_results


from typing import Optional
import uuid
import array
import numpy as np

from quickstats.interface.root import TObject, TArrayData
from quickstats.interface.root import load_macro
from quickstats.interface.cppyy.vectorize import as_np_array, vector_to_pointer

class TF1(TObject):
    
    def __init__(self, f:"ROOT.TF1"):
        self.f = f
        
    def GenerateRandomArray(self, size:int, xmin:Optional[float]=None, xmax:Optional[float]=None,
                            fmt:str="pointer"):
        import ROOT
        if (xmin is None) and (xmax is None):
            rand_arr = ROOT.TF1Utils.GetRandomArray(self.f, size)
        else:
            rand_arr = ROOT.TF1Utils.GetRandomArray(self.f, size, xmin, xmax)
        if fmt == "numpy":
            return as_np_array(rand_arr)
        elif fmt == "vector":
            return rand_arr
        elif fmt == "pointer":
            return vector_to_pointer(rand_arr)
        raise ValueError(f'invalid output format: {fmt}')


from typing import Optional, List

import numpy as np

from .rooproc_hybrid_action import RooProcHybridAction

from quickstats.utils.common_utils import is_valid_file

class RooProcAsNumpy(RooProcHybridAction):
    
    def __init__(self, filename:str, 
                 columns:Optional[List[str]]=None):
        super().__init__(filename=filename,
                         columns=columns)
        
    @classmethod
    def parse(cls, main_text:str, block_text:Optional[str]=None):
        kwargs = cls.parse_as_kwargs(main_text)
        return cls(**kwargs)
    
    def _execute(self, rdf:"ROOT.RDataFrame", processor:"quickstats.RooProcessor", **params):
        filename = kwargs['filename']
        if processor.cache and is_valid_file(filename):
            processor.stdout.info(f"INFO: Cached output `{filename}`.")
            return rdf, processor
        columns = kwargs.get('columns', None)
        if columns is None:
            data = rdf.AsNumpy()
        else:
            data = rdf.AsNumpy(columns)
        np.save(filename, data)
        return rdf, processor


def get_coeff_dict(expr, *x, strict:bool=True, linear:bool=False):
    if strict and (set(expr.free_symbols) != set(x)):
        raise RuntimeError("supplied symbols and expression symbols mismatch")
    from sympy import Poly, Add, S
    collected = Poly(expr, *x).as_expr()
    i, d = collected.as_independent(*x, as_Add=True)
    rv = dict(i.as_independent(*x, as_Mul=True)[::-1] for i in Add.make_args(d))
    if i:
        assert 1 not in rv
        rv.update({S.One: i})
    if linear and (not set(rv.keys()).issubset(set(x))):
        raise RuntimeError("expression contains nonlinear terms")
    return rv


from quickstats import ConfigComponent

DEFAULT_SAMPLE_CONFIG_FORMAT = {
    "sample_dir": ConfigComponent(dtypes="STR", default="./",
                                  description="directory in which the input root samples are located"),
    "sample_subdir": ConfigComponent(dtypes="DICT[STR]", default_factory=dict,
                                     description="(optional) a dictionary mapping the sample type to the sub-directory "
                                                 "of input root samples in which the input root samples are located; "
                                                 "examples of sample type are the mc campaigns and data year for "
                                                 "mc and data samples respectively"),
    "samples": ConfigComponent(dtypes=["DICT[DICT[STR]]", "DICT[DICT[LIST[STR]]]"], required=True,
                               description="A map from the sample name to the sample path in the form "
                                           "{<sample_type>: <sample_path>}; if path is given by an "
                                           "absolute path, the absolute path is used, otherwise the path "
                                           "{sample_dir}/{sample_subdir}/{path} is used",
                               example='"sample_name": {"ggF": {"mc16a": "ggF_mc16a.root", "mc16d": "ggF_mc16d.root"}}'),
    "merge_samples": ConfigComponent(dtypes="DICT[LIST[STR]]", default_factory=dict,
                                     description="merge the given list of samples into a sample with a given name in "
                                                 "the form {<merged_sample_name>: <list_of_samples_to_be_merged>}",
                                     example='"merge_samples": {"H": ["ggF", "VBF", "VH", "ttH"]}')
}

DEFAULT_ANALYSIS_CONFIG_FORMAT = {
    "paths": {
        "ntuples": ConfigComponent(dtypes="STR", default="./ntuples",
                                   description="path to analysis data in root format as precursor to analysis "
                                               "data arrays"),
        "arrays" : ConfigComponent(dtypes="STR", default="./arrays",
                                   description="path to analysis data in csv/h5 format used for "
                                               "model training and statistical analysis"),
        "outputs": ConfigComponent(dtypes="STR", default="./outputs",
                                   description="path to analysis outputs"),
        "models": ConfigComponent(dtypes="STR", default="./models",
                                  description="path to machine learning models"),
        "*": ConfigComponent(dtypes="STR",
                             description="any named path")
    },
    "samples": {
        "all": ConfigComponent(dtypes="LIST[STR]", default_factory=list,
                               description="list of all analysis samples"),
        "extra": ConfigComponent(dtypes="LIST[STR]", default_factory=list,
                                 description="list of additional analysis samples that exist only "
                                             "in the form of data arrays (i.e. not ntuples)"),
        "*"  : ConfigComponent(dtypes="LIST[STR]",
                               description="list of analysis samples belonging to a group with the given key",
                               example='"signal": ["ggH", "VBF", "VH", "ttH"]')
    },
    "kinematic_regions": ConfigComponent(dtypes="LIST[STR]", default_factory=list,
                                         description="kinematic regions of the analysis; if defined, input "
                                                     "datasets are assumed to be split according to these "
                                                     "regions; these regions typically correpond to the "
                                                     "relevant analysis channels",
                                         example='"kinematic_regions": ["zero_jet", "one_jet", "two_jet"]'),
    "variables":{
        "all": ConfigComponent(dtypes="LIST[STR]", default_factory=list,
                               description="list of all input variables"),
        "*"  : ConfigComponent(dtypes="LIST[STR]",
                               description="list of input variables belonging to a group with the given key",
                               example='"jets": ["jet1_pt", "jet1_eta", "jet2_pt", "jet2_eta"]')
    },
    "names": {
        "tree_name": ConfigComponent(dtypes="STR", default="output",
                                     description="tree name of ntuples used in the analysis"),
        "event_number": ConfigComponent(dtypes="STR", default="event_number",
                                        description="variable name that describes the event number of an event"),
        "*": ConfigComponent(dtypes="STR",
                             description="name of any key object in the analysis"),
    },
    "observable":{
        "name": ConfigComponent(dtypes="STR",
                                description="name of discriminant variable"),
        "eval_expr": ConfigComponent(dtypes=["STR", "NONE"], default=None,
                                     description="expression to evaluate to the discriminant from existing input variables"),
        "bin_range": ConfigComponent(dtypes="LIST[FLOAT]",
                                     description="bin range"),
        "blind_range": ConfigComponent(dtypes=["LIST[FLOAT]", "NONE"], default=None,
                                       description="blind range"),
        "n_bins": ConfigComponent(dtypes="INT",
                                  description="number of bins"),
    },    
    "training": {
        "datasets":{
            "specification": {
                "*": {
                    "selection": ConfigComponent(dtypes="STR",
                                                 description="selection applied to obtain the given dataset",
                                                 example='"training":{"specification":{"train_dataset": "event_number % 4 <= 2"'
                                                         ', "test_dataset": "event_number %4 == 3}})'),
                    "variables": ConfigComponent(dtypes="LIST[STR]", default_factory=list,
                                                 description="list of variables that should be loaded on top of the "
                                                             "training variables to perform the dataset selection"),
                    "effective_weight": ConfigComponent(dtypes="FLOAT", default=1,
                                                        description="effective weight of the dataset with respect "
                                                                    "to the inclusive data")
                }
            },
            "split_method": ConfigComponent(dtypes="STR", default="kfold",
                                            description="how to split the dataset for training; custom = use the "
                                                        "datasets defined in `specification`; kfold = split the "
                                                        "full dataset (or a dataset in `specification`) into k folds "
                                                        "; index = select data from a given set of indices (per-sample) "
                                                        "which can be the entry number or a column number;"
                                                        " manual = use a custom function to split data;",
                                            choice=["custom", "kfold", "index", "manual"]),
            "split_options": {
                "*": ConfigComponent(dtypes="ANY",
                                     description="option specific to a given split method")
            }
        },
        "sample_transformation":{
            "normalize_weight": ConfigComponent(dtypes="BOOL", default=False,
                                                description="whether to normalize the event weight in each sample")
        },
        "dataset_transformation":{
            "scale_weight_by_mean": ConfigComponent(dtypes="BOOL", default=False,
                                                    description="whether to rescale the event weight by the mean in"
                                                                 " the train/val/test datasets after samples are"
                                                                 " merged to the corresponding classes"),
            "negative_weight_mode": ConfigComponent(dtypes="INT", default=0,
                                                    description="how to treat negative weights: 0 = no action performed; "
                                                                "1 = set weight to 0; 2 = set weight to absolute value",
                                                    choice=[0, 1, 2]),
            "random_state": ConfigComponent(dtypes=["INT", "NONE"], default=-1,
                                            description="random state for shuffling data; if negative, "
                                                        "no shuffling will be made; if None, the global "
                                                        "random state instance from numpy.random will "
                                                        "be used (so every shuffle will give a different "
                                                        "result)")
        }
    },
    "categorization":{
        "boundary_scan":{
            "selection": ConfigComponent(dtypes="DICT[STR]", default_factory=dict,
                                         description="(per-sample) selection applied when performing boundary scan"),
            "datasets": ConfigComponent(dtypes=["LIST[STR]", "None"], default_factory=list,
                                        description="dataset(s) that should take non-zero weight when peroforming "
                                                    "boundary scan; this is typically the test dataset; if empty, "
                                                    "the full dataset is used"),
            "adaptive_dataset": ConfigComponent(dtypes="BOOL", default=True,
                                                description="use full dataset for samples not used in training and "
                                                            "the specified dataset for samples used in training")
        },
        "evaluation":{
            "datasets": ConfigComponent(dtypes=["LIST[STR]", "None"], default_factory=list,
                                        description="dataset(s) that should take the total weight when saving "
                                                    "the categorized outputs; this is typically the test dataset; "
                                                    "if empty, the full dataset is used"),
            "adaptive_dataset": ConfigComponent(dtypes="BOOL", default=True,
                                                description="use full dataset for samples not used in training and "
                                                            "the specified dataset for samples used in training")
        },
        "save_variables": ConfigComponent(dtypes="LIST[STR]", default_factory=list,
                                          description="additional variables to save in the categorized outputs"),
    },
    "data_storage":{
        "analysis_data_arrays": {
            "storage_format": ConfigComponent(dtypes="STR", default="csv",
                                              description="file format used to store the data",
                                              choice=["csv", "h5"]),
            "storage_options": ConfigComponent(dtypes="DICT", default_factory=dict,
                                               description="storage options specific to a given storage format",
                                               example='csv format: no storage options required; '
                                                       'h5 format: {"key": "analysis_data", "complevel": 5}')
        }
    },
    "channels":{
        "*":{
            "selection": ConfigComponent(dtypes=["STR", "NONE"], default=None,
                                         description="selection applied on the input variables to isolate the phase "
                                                     "space for the given channel",
                                         example='"channel": {"LowPtRegion": {"selection": "jet_pt < 125"}}'),
            "selection_variables": ConfigComponent(dtypes=["LIST[STR]", "NONE"], default=None,
                                                   description="variables used for applying the channel selection"), 
            "kinematic_region": ConfigComponent(dtypes=["STR", "NONE"], default=None,
                                                description="kinematic region corresponding to the given channel"),
            "train_samples": ConfigComponent(dtypes="LIST[STR]",
                                             description="training samples used for the given channel (group label is allowed)",
                                             example='"channel": {"LowPtRegion": {"train_samples": ["signal", "yj", "yy"]}}'),
            "test_samples": ConfigComponent(dtypes="LIST[STR]",
                                            description="test samples used for the given channel (group label is allowed); "
                                                        "categorized outputs will be produced for these samples",
                                            example='"channel": {"LowPtRegion": {"test_samples": ["all"]}}'),
            "train_variables": ConfigComponent(dtypes="LIST[STR]",
                                               description="training variables used for the given channel (group label is allowed)",
                                               example='"channel": {"LowPtRegion": {"train_variables": ["jets", "pt_H", "m_H"]}}'),
            "class_labels":ConfigComponent(dtypes="DICT[LIST[STR]]",
                                           description="a dictionary that maps the class label used in training "
                                                       "to the corresponding samples",
                                           example='"channel": {"LowPtRegion": {"class_labels": {"0": '
                                                   '["yj", "yy"], "1": ["signal"]}}}'),
            "hyperparameters": ConfigComponent(dtypes="DICT", default_factory=dict,
                                               description="a dictionary specifying the hyperparameters used in the training",
                                               example='"channel": {"LowPtRegion": {"hyperparameters": '
                                                       '{"learning_rate": 0.01, "batchsize": 100}}}'),
            "SF": ConfigComponent(dtypes="DICT", default_factory=dict,
                                  description="a dictionary that maps the scale factor applied to the weight "
                                              "of a sample used in the training",
                                  example='"channel": {"LowPtRegion": {"SF": {"ggF": 100, "VBF": 50}}}'),
            "counting_significance":{
                "?":{
                    "signal": ConfigComponent(dtypes="LIST[STR]",
                                              description="the samples designated as signals when evaluating the counting "
                                                          "significance in score boundary scans (group label is allowed)",
                                              example='"channel": {"LowPtRegion": {"counting_significance": '
                                                      '{"signal": ["ggF", "VBF"]}}}'),
                    "background": ConfigComponent(dtypes="LIST[STR]",
                                                  description="the samples designated as backgrounds when evaluating the "
                                                              "counting significance in score boundary scans (group label "
                                                              "is allowed)",
                                                  example='"channel": {"LowPtRegion": {"counting_significance": '
                                                          '{"background": ["yj", "yy"]}}}'),
                    "n_bins": ConfigComponent(dtypes="INT",
                                              description="Number of bins used in score boundary scan; notice the "
                                                          "scan time and memory consumption grow exponentially with "
                                                          "the number of bins used"),
                    "n_boundaries": ConfigComponent(dtypes="INT",
                                                    description="Number of score boundaries to apply; you will get "
                                                                "(n_boundaries + 1) categories for the given channel "
                                                                "if all categories are kept"),
                    "min_yield": ConfigComponent(dtypes="DICT[FLOAT]", default_factory=dict,
                                                 description="Minimum yield of specific samples required in all score regions",
                                                 example='"channel": {"LowPtRegion": {"counting_significance": '
                                                         '{"min_yield": {"yy": 2}}}}')
                }
            },
            "exclude_categories": ConfigComponent(dtypes="LIST[LIST[INT]]", default_factory=list,
                                                  description="Remove specific categories from the analysis by their "
                                                              "category index; [0] is the first bin of a 1D (binary "
                                                              "class) boundary scan, [0, 0] is the first bin of a 2D "
                                                              "(multiclass, e.g. [score_signal_1, score_signal_2]) "
                                                              "boundary scan",
                                                  example='"channel": {"LowPtRegion": {"exclude_categories": [[0]]')
        }
    }
}


from typing import Optional, Union, List, Dict, Tuple
import os

class DynamicFilePath:
    def __init__(self, basename:str, dirname:Optional[str]=None):
        self.dirname  = dirname
        self.basename = basename
        
    def resolve_basename(self, **parameters):
        return self.basename.format(**parameters)
        
class PathManager:
    
    @property
    def directories(self):
        return self._directories
    
    @directories.setter
    def directories(self, values:Optional[Dict[str, str]]=None):
        self._directories = self._parse_directories(values)
    
    @property
    def files(self):
        return self._files
    
    @files.setter
    def files(self, values:Optional[Union[Dict[str, Union[str, DynamicFilePath, Tuple[Optional[str], str]]],
                                    List[Union[str, DynamicFilePath, Tuple[Optional[str], str]]]]]=None):
        self._files = self._parse_files(values)
    
    def __init__(self, base_path:Optional[str]=None,
                 directories:Optional[Dict[str, str]]=None,
                 files:Optional[Union[Dict[str, Union[str, DynamicFilePath, Tuple[Optional[str], str]]],
                                List[Union[str, DynamicFilePath, Tuple[Optional[str], str]]]]]=None):
        self.base_path   = base_path
        self.directories = directories
        self.files       = files
    
    @staticmethod
    def _parse_directories(directories:Optional[Dict[str, str]]=None):
        if directories is None:
            new_directories = {}
        else:
            if not isinstance(directories, dict):
                raise TypeError("directories must be specified in dict format")
            for key, value in directories.items():
                if (not isinstance(key, str)) or (not isinstance(value, str)):
                    raise TypeError("directory key and value must be strings")
            # make a copy
            new_directories = {**directories}
        return new_directories
    
    @staticmethod
    def _parse_file_path(file_path:Union[str, DynamicFilePath, Tuple[Optional[str], str]]):
        if not isinstance(file_path, (str, tuple, DynamicFilePath)):
            raise TypeError("file path must be a tuple/string/DynamicFilePath")
        if isinstance(file_path, tuple):
            if len(file_path) != 2:
                raise ValueError("a tuple file path format must have a size of two in the form (dirname, basename)")
            return DynamicFilePath(file_path[1], file_path[0])
        elif isinstance(file_path, str):
            return DynamicFilePath(file_path, None)
        elif isinstance(file_path, DynamicFilePath):
            return file_path
        else:
            raise TypeError("unknown file path format")    
    
    @staticmethod
    def _parse_files(files:Optional[Union[Dict[str, Union[str, DynamicFilePath, Tuple[Optional[str], str]]],
                                    List[Union[str, DynamicFilePath, Tuple[Optional[str], str]]]]]=None):
        if files is None:
            new_files = {}
        else:
            if isinstance(files, str):
                files = [files]
            if isinstance(files, list):
                files = {str(i):value for i, value in enumerate(files)}
            if not isinstance(files, dict):
                raise TypeError("files must be specified in dict format")
            new_files = {}
            for key, value in files.items():
                if not isinstance(key, str):
                    raise TypeError("file name must be a string")
                file = PathManager._parse_file_path(value)
                new_files[key] = file
        return new_files
        
    def update_directories(self, directories:Optional[Dict[str, str]]=None):
        new_directories = self._parse_directories(directories)
        self._directories.update(new_directories)
        
    def update_files(self, files:Optional[Dict[str, Union[str, DynamicFilePath, Tuple[Optional[str], str]]]]=None):
        new_files = self._parse_files(files)
        self._files.update(new_files)
        
    def set_directory(self, directory_name:str, path:str, absolute:bool=False):
        if absolute:
            path = os.path.abspath(path)
        self.update_directories({directory_name: path})
        
    def set_file(self, file_name:str, file:Union[str, DynamicFilePath, Tuple[Optional[str], str]]):
        self.update_files({file_name: file})
        
    def get_directory(self, directory_name:str, check_exist:bool=False):
        if directory_name not in self.directories:
            raise KeyError(f'unrecognized directory name "{directory_name}"')
        if self.base_path is not None:
            return os.path.join(self.base_path, self.directories[directory_name])
        directory = self.directories[directory_name]
        if check_exist:
            if not os.path.exists(directory):
                raise FileNotFoundError(f'directory "{directory}" does not exist')
        return directory
    
    def get_resolved_file(self, file_name:str, check_exist:bool=False,
                          subdirectory:Optional[str]=None, **parameters):
        if file_name not in self.files:
            raise KeyError(f'unrecognized file name "{file_name}"')
        file = self.files[file_name]
        if parameters:
            basename = file.resolve_basename(**parameters)
        else:
            basename = file.basename
        if subdirectory is None:
            subdirectory = ""
        if file.dirname is not None:
            dirname  = self.get_directory(file.dirname)
            fullname = os.path.join(dirname, subdirectory, basename)
        else:
            if self.base_path is not None:
                fullname = os.path.join(self.base_path, subdirectory, basename)
            else:
                fullname = os.path.join(subdirectory, basename)
        if check_exist:
            if not os.path.exists(fullname):
                raise FileNotFoundError(f'file "{fullname}" does not exist')
        return fullname
    
    def get_file(self, file_name:str, check_exist:bool=False):
        return self.get_resolved_file(file_name, check_exist=check_exist)
    
    def get_directories(self, directory_names:Optional[List[str]]=None):
        directory_paths = {}
        if directory_names is None:
            directory_names = list(self.directories.keys())
        for directory_name in directory_names:
            directory_paths[directory_name] = self.get_directory(directory_name)
        return directory_paths
    
    def get_files(self, file_names:Optional[List[str]]=None):
        file_paths = {}
        if file_names is None:
            file_names = list(self.files.keys())
        for file_name in file_names:
            file_paths[file_name] = self.get_file(file_name)
        return file_paths
    
    def get_relpath(self, path:str):
        if self.base_path is None:
            return path
        return os.path.join(self.base_path, path)
    
    def directory_exists(self, directory_name:str):
        directory = self.get_directory(directory_name)
        return os.path.exists(directory)
    
    def file_exists(self, file_name:str, **parameters):
        file = self.get_resolved_file(file_name, **parameters)
        return os.path.exists(file)
    
    def check_directory(self, directory_name:str):
        self.get_directory(directory_name, check_exist=True)
            
    def check_file(self, file_name:str, **parameters):
        self.get_resolved_file(file_name, check_exist=True, **parameters)
    
    def makedirs(self, include_names:Optional[List[str]]=None,
                 exclude_names:Optional[List[str]]=None):
        if include_names is None:
            include_names = list(self.directories.keys())
        if exclude_names is None:
            exclude_names = []
        directory_names = list(set(include_names) - set(exclude_names))
        for directory_name in directory_names:
            if directory_name not in self.directories:
                raise KeyError(f'unrecognized directory name "{directory_name}"')
            dirname = self.get_directory(directory_name)
            if not os.path.exists(dirname):
                os.makedirs(dirname)
    
    def makedir_for_files(self, file_names:Union[str, List[str]]):
        if isinstance(file_names, str):
            file_names = [file_names]
        files = self.get_files(file_names)
        for file in files.values():
            dirname = os.path.dirname(file)
            if (dirname != "") and not os.path.exists(dirname):
                os.makedirs(dirname)


from typing import Optional, Union, Dict, List, Tuple
import os

from quickstats import PathManager
from quickstats.utils.common_utils import combine_dict

class AnalysisPathManager(PathManager):
    
    DEFAULT_DIRECTORIES = {
        "categorized_array"     : "categorized_arrays",
        "categorized_minitree"  : "categorized_minitrees",
        "categorized_histogram" : "categorized_histograms",
        "yield"                 : "yields",
        "summary"               : "summary",
        "plot"                  : "plots",
        "xml"                   : "xmls",
        "workspace"             : "xmls/workspace",
        "xml_config"            : "xmls/config",
        "xml_data"              : "xmls/config/data",
        "xml_model"             : "xmls/config/models",
        "xml_category"          : "xmls/config/categories",
        "limit"                 : "limits",
        "likelihood"            : "likelihoods"
    }
    
    DEFAULT_FILES = {
        "ntuple_sample"                            : ("ntuple", "{sample}.root"),
        "train_sample"                             : ("array", "{sample}.{fmt}"),
        "array_sample"                             : ("array", "{sample}.{fmt}"),
        "categorized_array_sample"                 : ("categorized_array", "{sample}_{category}.{fmt}"),
        "categorized_minitree_sample"              : ("categorized_minitree", "{sample}_{category}.root"),
        "categorized_histogram_sample"             : ("categorized_histogram", "{sample}_{category}.root"),
        "category_summary"                         : ("summary", "category_summary_{channel}.json"),
        "boundary_data"                            : ("summary", "boundary_tree_{channel}.json"),
        "score_distribution_plot"                  : ("plot", "score_distribution_{channel}.pdf"),
        "variable_distribution_plot"               : ("plot", "distribution_{variable}_{category}.pdf"),   
        "merged_yield_data"                        : ("yield", "yields.json"),
        "merged_yield_err_data"                    : ("yield", "yields_err.json"),
        "yield_data"                               : ("yield", "yields_{category}.json"),
        "yield_err_data"                           : ("yield", "yields_err_{category}.json"),  
        "signal_modelling_plot"                    : ("plot", "modelling_{category}.pdf"),
        "signal_modelling_data"                    : ("signal_model", "model_parameters.json"),
        "signal_modelling_summary"                 : ("signal_model", "model_summary_{category}.json"),
        "input_xml"                                : ("xml_config", "input.xml"),
        "limit_summary"                            : ("summary", "limit_summary.json"),
        "likelihood_summary"                       : ("summary", "likelihood_summary.json"),
        "benchmark_significance"                   : ("summary", "benchmark_significance.json")
    }
    
    def __init__(self, study_name:str="", base_path:Optional[str]=None,
                 directories:Optional[Dict[str, str]]=None,
                 files:Optional[Dict[str, Union[str, Tuple[Optional[str], str]]]]=None):
        
        directories = combine_dict(self.DEFAULT_DIRECTORIES, directories)
        files       = combine_dict(self.DEFAULT_FILES, files)
            
        super().__init__(base_path=base_path, directories=directories, files=files)
        
        self.study_name    = study_name
        self.raw_base_path = base_path
        self.update()
        
    def update(self):
        if self.raw_base_path is None:
            base_path = self.study_name
        else:
            base_path = os.path.join(self.raw_base_path, self.study_name)
        self.base_path = base_path
        
    def set_study_name(self, study_name:str):
        self.study_name = study_name
        self.update()
        
    def set_base_path(self, base_path:str):
        self.raw_base_path = base_path
        self.update()


from quickstats.components.processors.actions import *
from quickstats.components.processors.roo_config_parser import RooConfigParser
from quickstats.components.processors.roo_processor import RooProcessor


from .component_source import ComponentSource
from .tree_data_source import TreeDataSource
from .model_source import ModelSource
from .pdf_fit_tool import PdfFitTool
from .data_modelling import DataModelling


from typing import Optional, Dict, List, Union
import copy

from quickstats import GeneralEnum, DescriptiveEnum
from quickstats.components import AnalysisObject
from quickstats.utils.common_utils import parse_config
from quickstats.components.basics import WSArgument

class AsimovType(DescriptiveEnum):
    Temp                         = (-999, "do not generate any asimov")
    S_NP_Nom                     = (-2, "S + B Asimov (mu = 1) with nominal NP values")
    B_NP_Nom                     = (-1, "B only Asimov (mu = 0) with nominal NP values")
    B_NP_Fit                     = (0, "B only Asimov (mu = 0) with NP values profiled to data (fit with mu fixed at 0)")
    S_NP_Fit                     = (1, "S + B Asimov (mu = 1) with NP values profiled to data (fit with mu fixed at 1)")
    S_NP_Fit_muhat               = (2, "S + B Asimov (mu = 1) with NP values profiled to data (fit with mu floating)")
    B_unconstrained_NP_Fit       = (3, "B only Asimov (mu = 0) with unconstrained NP values profiled to data "
                                   "(fit with mu fixed at 0 and constrained NP fixed at 0)")
    S_unconstrained_NP_Fit       = (4, "S + B Asimov (mu = 1) with unconstrained NP values profiled to data "
                                   "(fit with mu fixed at 1 and constrained NP fixed at 0)")
    S_unconstrained_NP_Fit_muhat = (5, "S + B Asimov (mu = 1) with unconstrained NP values profiled to data "
                                   "(fit with mu floating and constrained NP fixed at 0)")
    
class AsimovGenerator(AnalysisObject):

    ASIMOV_SETTINGS = {
        AsimovType.Temp: None,
        AsimovType.S_NP_Nom: {
            "asimov_name": "asimovData_1_NP_Nominal",
            "asimov_snapshot": "asimovData_1_NP_Nominal",
            "poi_val": 1,
            "poi_profile": 1,
            "do_fit": False,
            "modify_globs": False
        },
        AsimovType.B_NP_Nom: {
            "asimov_name": "asimovData_0_NP_Nominal",
            "asimov_snapshot": "asimovData_0_NP_Nominal",
            "poi_val": 0,
            "poi_profile": 0,
            "do_fit": False,
            "modify_globs": False
        },
        AsimovType.B_NP_Fit: {
            "asimov_name": "asimovData_0_NP_Profile",
            "asimov_snapshot": "asimovData_0_NP_Profile",
            "poi_val": 0,
            "poi_profile": 0,
            "do_fit": True,
            "modify_globs": True
        },
        AsimovType.S_NP_Fit: {
            "asimov_name": "asimovData_1_NP_Profile",
            "asimov_snapshot": "asimovData_1_NP_Profile",
            "poi_val": 1,
            "poi_profile": 1,
            "do_fit": True,
            "modify_globs": True
        },
        AsimovType.S_NP_Fit_muhat: {
            "asimov_name": "asimovData_muhat_NP_Profile",
            "asimov_snapshot": "asimovData_muhat_NP_Profile",
            "poi_val": 1,
            "poi_profile": None,
            "do_fit": True,
            "modify_globs": True
        },
        AsimovType.B_unconstrained_NP_Fit: {
            "asimov_name": "asimovData_0_unconstrained_NP_Profile",
            "asimov_snapshot": "asimovData_0_unconstrained_NP_Profile",
            "poi_val": 0,
            "poi_profile": 0,
            "do_fit": True,
            "modify_globs": True,
            "constraint_option": 1
        },
        AsimovType.S_unconstrained_NP_Fit: {
            "asimov_name": "asimovData_1_unconstrained_NP_Profile",
            "asimov_snapshot": "asimovData_1_unconstrained_NP_Profile",
            "poi_val": 1,
            "poi_profile": 1,
            "do_fit": True,
            "modify_globs": True,
            "constraint_option": 1
        },
        AsimovType.S_unconstrained_NP_Fit_muhat: {
            "asimov_name": "asimovData_muhat_unconstrained_NP_Profile",
            "asimov_snapshot": "asimovData_muhat_unconstrained_NP_Profile",
            "poi_val": 1,
            "poi_profile": None,
            "do_fit": True,
            "modify_globs": True,
            "constraint_option": 1
        }
    }    
    
    DEFAULT_ASIMOV_TYPES = [AsimovType.B_NP_Fit, AsimovType.S_NP_Fit, AsimovType.S_NP_Fit_muhat]
    
    def __init__(self, filename:str, poi_name:str=None,
                 data_name:str='combData', 
                 config:Optional[Union[Dict, str]]=None,
                 verbosity:Optional[Union[int, str]]="INFO"):
        config = parse_config(config)
        config['filename']  = filename
        config['poi_name']  = poi_name
        config['data_name'] = data_name
        config['verbosity'] = verbosity
        self._inherit_init(super(AsimovGenerator, self).__init__, **config)
        
    def generate_asimov(self, poi_val:Optional[float]=None, 
                        poi_profile:Optional[float]=None,
                        do_fit:bool=True,
                        modify_globs:bool=True,
                        do_import:bool=True,
                        asimov_name:Optional[str]=None,
                        asimov_snapshot:Optional[str]=None,
                        channel_asimov_name:Optional[str]=None,
                        dataset:Optional["ROOT.RooDataSet"]=None,
                        constraint_option:int=0,
                        restore_states:int=0,
                        snapshot_names:Optional[Dict]=None):
        kwargs = {
            "poi_name"            : self.poi.GetName(),
            "poi_val"             : poi_val,
            "poi_profile"         : poi_profile,
            "do_fit"              : do_fit,
            "modify_globs"        : modify_globs,
            "do_import"           : do_import,
            "asimov_name"         : asimov_name,
            "asimov_snapshot"     : asimov_snapshot,
            "channel_asimov_name" : channel_asimov_name,
            "dataset"             : dataset,
            "constraint_option"   : constraint_option,
            "restore_states"      : restore_states,
            "minimizer_options"   : self.minimizer_options,
            "nll_options"         : self.nll_commands,
            "snapshot_names"      : snapshot_names
        }
        return self.model.generate_asimov(**kwargs)
        
    def generate_standard_asimov(self, asimov_types:Optional[Union[List[AsimovType], List[int], List[str]]]=None,
                                 asimov_names:Optional[Union[List[str], str]]=None,
                                 asimov_snapshots:Optional[Union[List[str], str]]=None,
                                 poi_scale:Optional[float]=None, do_import:bool=True):
        single_asimov = False
        if asimov_types is None:
            asimov_types = self.DEFAULT_ASIMOV_TYPES
        if isinstance(asimov_types, (int, str, AsimovType)):
            asimov_types = [AsimovType.parse(asimov_types)]
            single_asimov = True
        else:
            asimov_types = [AsimovType.parse(atype) for atype in asimov_types]
        if asimov_names is not None:
            if isinstance(asimov_names, str):
                asimov_names = [asimov_names]
            assert len(asimov_names) == len(asimov_types)
        if asimov_snapshots is not None:
            if isinstance(asimov_snapshots, str):
                asimov_snapshots = [asimov_snapshots]
            assert len(asimov_snapshots) == len(asimov_types)            
        if poi_scale is None:
            poi_scale = 1.0
        poi_name = self.poi.GetName()
        self.save_snapshot("temp", WSArgument.MUTABLE)
        asimov_datasets = []
        for i, asimov_type in enumerate(asimov_types):
            self.load_snapshot("temp")
            kwargs = copy.deepcopy(self.ASIMOV_SETTINGS[asimov_type])
            if kwargs is None:
                continue
            kwargs['poi_name'] = poi_name
            for key in ['poi_val', 'poi_profile']:
                if kwargs[key] is not None:
                    kwargs[key] *= poi_scale
            if asimov_names is not None:
                kwargs['asimov_name'] = asimov_names[i]
            if asimov_snapshots is not None:
                kwargs['asimov_snapshot'] = asimov_snapshots[i]
            kwargs['minimizer_options'] = self.minimizer.config
            kwargs['nll_options'] = list(self.minimizer.nll_commands.values())
            kwargs['do_import'] = do_import
            asimov_dataset = self.model.generate_asimov(**kwargs)
            asimov_datasets.append(asimov_dataset)
        if single_asimov:
            return asimov_datasets[0]
        return asimov_datasets


from typing import List

from .rooproc_nested_action import RooProcBaseAction, RooProcNestedAction
from .auxiliary import RooProcReturnCode

class RooProcIfDefined(RooProcNestedAction):
    
    def __init__(self, flag:str):
        super().__init__(flag=flag)
        
    @classmethod
    def parse(cls, main_text:str, block_text:str):
        if not block_text:
            raise ValueError("missing flag name in IFDEF action")
        return cls(flag=block_text)
     
    def _execute(self, processor:"quickstats.RooProcessor", **params):
        flag = params['flag']
        if flag in processor.flags:
            return RooProcReturnCode.NORMAL
        return RooProcReturnCode.SKIP_CHILD


from typing import Union, Optional, List, Tuple, Dict
import copy
import json

from quickstats import semistaticmethod

def int_keys(ordered_pairs):
    result = {}
    for key, value in ordered_pairs.items():
        try:
            key = int(key)
        except ValueError:
            pass
        result[key] = value
    return result

class MultiClassBoundaryTree:
    def __init__(self, score_base_name:str="score"):
        self.tree = {}
        self.current_node = None
        self.score_base_name = score_base_name
    
    @semistaticmethod
    def iter_branch(self, branch):
        if not branch:
            return {}
        sub_branch_keys = branch['branch']
        for key in sub_branch_keys:
            sub_branch = branch['branch'][key]
            if not isinstance(key, int):
                branch['branch'].pop(key)
            branch['branch'][int(key)] = self.iter_branch(sub_branch)
        return branch
    
    def load_tree(self, filename:str):
        with open(filename, "r") as file:
            self.tree = json.load(file, object_hook=int_keys)
        
    def get_branch(self, branch_index:Optional[Union[Tuple[int], int]]=None):
        branch = self.tree
        if branch_index is not None:
            if isinstance(branch_index, int):
                branch_index = [branch_index]
            for index in branch_index:
                if 'branch' not in branch:
                    raise ValueError("tree level out of bounds")
                if index not in branch['branch']:
                    raise ValueError("tree index out of bounds")
                branch = branch['branch'][index]
        return branch   
    
    def get_boundaries(self, branch_index:Optional[Union[Tuple[int], int]]=None):
        branch = self.get_branch(branch_index)
        if 'boundaries' not in branch:
            raise RuntimeError("boundaries not defined")
        boundaries = branch['boundaries']
        return boundaries
    
    def set_boundaries(self, class_label:str, boundaries:List[float],
                       branch_index:Optional[Union[Tuple[int], int]]=None, **kwargs):
        leaf = {
            'class': class_label,
            'boundaries': boundaries,
            'branch': {i:{} for i in range(len(boundaries)+1)},
            **kwargs
        }
        target_branch = self.get_branch(branch_index)
        target_branch.update(leaf)      
        
    @staticmethod
    def _get_score_windows(boundaries:List[float]):
        score_windows = []
        closed_boundaries = [0.] + boundaries + [1.]
        for i in range(len(closed_boundaries)-1):
            score_windows.append([closed_boundaries[i], closed_boundaries[i+1]])
        return score_windows

    def get_branch_indices(self, current_branch:Optional[Dict]=None,
                           current_index:Optional[List[int]]=None):
        output = []
        if current_branch is None:
            current_branch = self.tree
        if "branch" in current_branch:
            for branch_index in current_branch["branch"]:
                next_branch = current_branch["branch"][branch_index]
                if current_index is None:
                    next_index = [branch_index]
                    
                else:
                    next_index = list(current_index) + [branch_index]
                output.append(next_index)
                output += self.get_branch_indices(next_branch, next_index)
        return output    
    
    def get_combined_boundaries(self, branch_index:Union[int, List[int]]):
        branch = self.tree
        combined_boundaries = {}
        if isinstance(branch_index, int):
            branch_index = [branch_index]
        for index in branch_index:
            class_name = branch['class']
            boundaries = branch['boundaries']
            score_windows = self._get_score_windows(boundaries)
            # binary-class case
            if (not class_name) and len(branch_index) == 1:
                return score_windows[index]
            if class_name in combined_boundaries:
                raise RuntimeError("found multiple classes with the same name")
            combined_boundaries[class_name] = score_windows[index]
            if 'branch' not in branch:
                raise ValueError("tree level out of bounds")
            if index not in branch['branch']:
                raise ValueError("tree index out of bounds")
            branch = branch['branch'][index]
        return combined_boundaries

    def get_cut_maps(self, source:Dict=None, result:Optional[dict]=None,
                     index:Optional[List[int]]=None, 
                     cuts:Optional[List[str]]=None):
        if source is None:
            source = self.tree
            if not source:
                return {None: None}
        if result is None:
            result = {}
        if index is None:
            index = []
        if cuts is None:
            cuts = []
        class_name = source['class']
        boundaries = source['boundaries']
        score_windows = self._get_score_windows(boundaries)
        if not class_name:
            score_label = self.score_base_name
        else:
            score_label = f"{class_name}_{self.score_base_name}"
        for i, score_window in enumerate(score_windows):
            if score_window[1] != 1.:
                cut = f"({score_label} >= {score_window[0]}) & ({score_label} < {score_window[1]})"
            else:
                cut = f"({score_label} >= {score_window[0]}) & ({score_label} <= {score_window[1]})"
            index_i = list(index) + [i]
            cuts_i = list(cuts) + [cut]
            branch = source['branch'][i]
            if not branch:
                result[tuple(index_i)] = " & ".join(cuts_i)
            else:
                self.get_cut_maps(branch, result, index_i, cuts_i)
        return result
    
    @staticmethod
    def iter_branch_sig(branch:Dict, sublabel:str=None, temp_result:Optional[Dict]=None):
        if temp_result is None:
            temp_result = {}
        class_name = branch['class']
        significance = branch['significance']
        if class_name not in temp_result:
            temp_result[class_name] = {}
        if sublabel is not None:
            temp_result[class_name][sublabel] = significance
        else:
            temp_result[class_name] = significance
        for branch_index in branch['branch']:
            sub_branch = branch['branch'][branch_index]
            if sub_branch:
                if sublabel is None:
                    _sublabel = f"{class_name}_{branch_index}"
                else:
                    _sublabel = f"{sublabel}_{class_name}_{branch_index}"            
                MultiClassBoundaryTree.iter_branch_sig(sub_branch, _sublabel, temp_result)
        return temp_result
    
    def get_significance_summary(self):
        summary = self.iter_branch_sig(self.tree)
        # just report the significance value in case of binary class
        if (len(summary) == 1) and (("" in summary) or None in summary):
            key = list(summary)[0]
            return summary[key]
        return summary
    
    @semistaticmethod
    def _get_str_repr(self, source:Dict, level:int=0):
        str_repr = ""
        class_name = source['class']
        if class_name is None:
            class_name = ""
        boundaries = source['boundaries']
        score_windows = self._get_score_windows(boundaries)
        if 'significance' in source:
            extra_txt = f" (Z = {source['significance']:.3f})"
        else:
            extra_txt = ""
        str_repr += ("  "*2*level + class_name + extra_txt + "\n")
        for i, score_window in enumerate(score_windows):
            str_repr += ("  "*(2*level+1) + f"{i}: [{score_window[0]}, {score_window[1]}]" + "\n")
            branch = source['branch'][i]
            if not branch:
                continue
            str_repr += self._get_str_repr(branch, level+1)
        return str_repr
    
    def __repr__(self):
        return self._get_str_repr(self.tree)


from typing import List, Dict, Union, Optional
import os
import copy
from itertools import chain
from dataclasses import dataclass

import pandas as pd
import numpy as np

from quickstats import (semistaticmethod, DescriptiveEnum, DataclassObject,
                        ConfigComponent, ConfigParser, ConfigurableObject)
from .analysis_base import AnalysisBase
from .data_preprocessing import fix_negative_weights, shuffle_arrays, get_class_label_encodings

@dataclass
class SampleTransformation(DataclassObject):
    normalize_weight: bool = False

@dataclass
class DatasetTransformation(DataclassObject):
    scale_weight_by_mean: bool = False
    negative_weight_mode: int = 0
    random_state: int = None 

class DataTransformer(object):
    
    def __init__(self, sample_weight_scales:Optional[Dict]=None,
                 sample_transformation:Optional[Dict]=None,
                 dataset_transformation:Optional[Dict]=None,
                 **kwargs):
        self.set_sample_weight_scales(sample_weight_scales)
        self.sample_transformation  = SampleTransformation.from_dict(sample_transformation)
        self.dataset_transformation = DatasetTransformation.from_dict(dataset_transformation)
        
    def set_random_state(self, random_state:Optional[int]=-1):
        self.dataset_transformation.random_state = random_state
        
    @semistaticmethod
    def make_copy(self, x:pd.DataFrame, y:pd.DataFrame,
                  weight:Optional[pd.DataFrame]=None):
        x_copy = x.copy()
        y_copy = y.copy()
        if weight is not None:
            weight_copy = weight.copy()
        else:
            weight_copy = None
        return x_copy, y_copy, weight_copy
    
    def transform_sample_data(self, x:pd.DataFrame, y:pd.DataFrame,
                              weight:Optional[pd.DataFrame]=None,
                              sample:Optional[str]=None,
                              **kwargs):
        x, y, weight = self.make_copy(x, y, weight)
        if weight is not None:
            if self.sample_transformation.normalize_weight:
                weight["weight"] = weight["weight"] / weight["weight"].sum()
            if sample in self.sample_weight_scales:
                weight_scale = self.sample_weight_scales[sample]
                weight["weight"] = weight["weight"] * weight_scale
        return x, y, weight
    
    def transform_dataset_data(self, x:pd.DataFrame, y:pd.DataFrame,
                               weight:Optional[pd.DataFrame]=None,
                               **kwargs):
        x, y, weight = self.make_copy(x, y, weight)      
        if weight is not None:
            if self.dataset_transformation.scale_weight_by_mean:
                weight["weight"] = weight["weight"] / weight['weight'].mean()
            mode = self.dataset_transformation.negative_weight_mode
            fix_negative_weights(weight, mode=mode, weight_col="weight")
        random_state = self.dataset_transformation.random_state
        if weight is None:
            x, y = shuffle_arrays(x, y, random_state=random_state)
        else:
            x, y, weight = shuffle_arrays(x, y, weight, random_state=random_state)
        return x, y, weight
    
    def transform_and_merge(self, sample_data:Dict, **kwargs):
        transformed_data = {}
        for sample, data in sample_data.items():
            x, y, weight = self.transform_sample_data(**data, sample=sample, **kwargs)
            transformed_data[sample] = {"x": x, "y": y, "weight": weight}
        else:
            has_weight = weight is not None
        # concatenate samples
        x      = pd.concat([transformed_data[sample]["x"] for sample in transformed_data], ignore_index=True)
        y      = pd.concat([transformed_data[sample]["y"] for sample in transformed_data], ignore_index=True)
        if has_weight:
            weight = pd.concat([transformed_data[sample]["weight"] for sample in transformed_data], ignore_index=True)
        else:
            weight = None
        x, y, weight = self.transform_dataset_data(x, y, weight, **kwargs)
        return x, y, weight
                  
    def set_sample_weight_scales(self, sample_weight_scales:Optional[Dict]=None):
        if sample_weight_scales is None:
            self.sample_weight_scales = {}
        else:
            self.sample_weight_scales = sample_weight_scales

class SplitAlgoBase(ConfigurableObject):
    def get_analysis_dataset_names(self) -> List[str]:
        # use the full dataset by default
        return None
        
    def get_extra_variables(self):
        return []
    
    def train_val_test_split(self, analysis_datasets:Dict,
                             data_transformer:DataTransformer,
                             **kwargs):
        raise NotImplementedError
        
    def set_random_state(self, random_state:Optional[int]=-1):
        self.config["random_state"] = random_state        
                  
    def combine_analysis_datasets(self, analysis_datasets:List[Dict]):
        combined_dataset = {}
        for dataset in analysis_datasets:
            for sample in dataset:
                if sample not in combined_dataset:
                    combined_dataset[sample] = dataset[sample].copy()
                    continue
                combined_dataset[sample]["x"] = pd.concat([combined_dataset[sample]["x"], dataset[sample]["x"]],
                                                           ignore_index=True)
                combined_dataset[sample]["y"] = pd.concat([combined_dataset[sample]["y"], dataset[sample]["y"]],
                                                           ignore_index=True)
                if dataset[sample]["weight"] is None:
                    continue
                combined_dataset[sample]["weight"] = pd.concat([combined_dataset[sample]["weight"], dataset[sample]["weight"]],
                                                               ignore_index=True)
        return combined_dataset                  
        
    def get_dataset_split(self, data_transformer:DataTransformer,
                         dataset_type:str, sample_data:Dict,
                         sample_index:Optional[Dict]=None):
        x, y, weight = data_transformer.transform_and_merge(sample_data)
        extra_variables = self.get_extra_variables()
        if len(extra_variables) > 0:
            x = x.drop(columns=extra_variables)
        dataset_split = {
            f"x_{dataset_type}"      : x,
            f"y_{dataset_type}"      : y,
            f"weight_{dataset_type}" : weight,
        }
        if sample_index is not None:
            dataset_split[f"index_{dataset_type}"] = sample_index
        return dataset_split

class KFoldSplitAlgo(SplitAlgoBase):
    
    CONFIG_FORMAT = {
        "datasets": ConfigComponent(dtypes=["LIST[STR]", "NONE"], default=None,
                                    description="dataset(s) that should be used in kfold splitting; if none, the "
                                                "full dataset will be used"),
        "index_variable": ConfigComponent(dtypes=["STR", "NONE"], default=None,
                                          description="the variable used for indexing (to keep track of "
                                                      "train/val events in each fold); if none, "
                                                      "no indices will be saved"),
        "n_splits": ConfigComponent(dtypes="INT", required=True,
                                    description="number of splits (= number of folds); one of the split "
                                                "will be used for validation and the rest for training "
                                                "with each fold rotating the splits"),
        "random_state": ConfigComponent(dtypes=["INT", "NONE"], default=-1,
                                        description="random state for shuffling data in each fold; "
                                                    "if negative, no shuffling will be made; "
                                                    "if None, the global random state instance from "
                                                    "numpy.random will be used (so every shuffle "
                                                    "will give a different result)")
    }
    
    def get_analysis_dataset_names(self) -> List[str]:
        dataset_config = self.config.get("datasets", None)
        if dataset_config is not None:
            dataset_names = list(dataset_config)
        else:
            dataset_names = None
        return dataset_names
    
    def get_extra_variables(self):
        index_variable = self.config.get("index_variable", None)
        if index_variable is not None:
            return [index_variable]
        return []
    
    def train_val_test_split(self, analysis_datasets:Dict,
                             data_transformer:DataTransformer,
                             **kwargs):
        
        n_splits     = self.config["n_splits"]
        random_state = self.config["random_state"]
        
        self.stdout.info("INFO: Data will be split using the KFold method.")
        self.stdout.info(f"   Analysis datasets used: {self.get_analysis_dataset_names()}")
        self.stdout.info(f"         Number of splits: {n_splits}")
        self.stdout.info(f"       KFold random state: {random_state}")
        
        from sklearn.model_selection import KFold
        
        index_variable = self.config.get("index_variable", None)
        kfolds = {}
        kfold_splits = {}
        combined_dataset = self.combine_analysis_datasets(list(analysis_datasets.values()))
        random_state = self.config["random_state"]
        if (random_state is not None) and (random_state < 0):
             shuffle = False
             random_state = None
        else:
             shuffle = True
        for sample in combined_dataset:
            kfolds[sample] = KFold(n_splits=n_splits,
                                   shuffle=shuffle,
                                   random_state=random_state)
            kfold_splits[sample] = kfolds[sample].split(combined_dataset[sample]["x"])
        for i in range(n_splits):
            train_dataset = {}
            val_dataset = {}
            train_var_index = {}
            val_var_index = {}
            for sample in combined_dataset:
                train_index, val_index = next(kfold_splits[sample])
                for index, dataset_i, var_index in [(train_index, train_dataset, train_var_index),
                                                    (val_index, val_dataset, val_var_index)]:
                    dataset_i[sample] = {
                        "x": combined_dataset[sample]["x"].loc[index],
                        "y": combined_dataset[sample]["y"].loc[index]
                    }
                    if combined_dataset[sample]["weight"] is not None:
                        dataset_i[sample]["weight"] = combined_dataset[sample]["weight"].loc[index]
                    if index_variable is not None:
                        var_index[sample] = dataset_i[sample]["x"][index_variable].values
            result_i = {}
            for dataset_type, dataset_i, var_index in [("train", train_dataset, train_var_index),
                                                       ("val", val_dataset, val_var_index)]:
                
                result_i.update(self.get_dataset_split(data_transformer, dataset_type, dataset_i, var_index))
            yield result_i
    
class CustomSplitAlgo(SplitAlgoBase):
    
    CONFIG_FORMAT = {
        "datasets": {
            "train": ConfigComponent(dtypes="LIST[STR]", required=True,
                                     description="dataset(s) that should be used as the train dataset"),
            "validation": ConfigComponent(dtypes=["LIST[STR]", "NONE"], default=None,
                                          description="dataset(s) that should be used as the validation dataset; "
                                                      "if none, no validation dataset will be loaded"),
            "test": ConfigComponent(dtypes=["LIST[STR]", "NONE"], default=None,
                                    description="dataset(s) that should be used as the test dataset; "
                                                "if none, no test dataset will be loaded")
        }
    }
    
    def get_train_val_test_map(self):
        datasets_config = self.config["datasets"]
        dataset_map = {}
        dataset_map["train"] = list(datasets_config["train"])
        if "validation" in datasets_config:
            dataset_map["val"] = list(datasets_config["validation"])
        if "test" in datasets_config:
            dataset_map["test"] = list(datasets_config["test"])
        return dataset_map
    
    def get_analysis_dataset_names(self) -> List[str]:
        dataset_map = self.get_train_val_test_map()
        dataset_names = list(chain.from_iterable(dataset_map.values()))
        return dataset_names
    
    def train_val_test_split(self, analysis_datasets:Dict, data_transformer:DataTransformer):
        dataset_map = self.get_train_val_test_map()
        self.stdout.info("INFO: Data will be split using custom analysis datasets.")
        self.stdout.info(f'        Train dataset: {dataset_map["train"]}')
        self.stdout.info(f'   Validation dataset: {dataset_map.get("val", None)}')
        self.stdout.info(f'         Test dataset: {dataset_map.get("test", None)}')
        result = {}
        for dataset_type, dataset_names in dataset_map.items():
            selected_datasets = [analysis_datasets[dataset_name] for dataset_name in dataset_names]
            combined_dataset = self.combine_analysis_datasets(selected_datasets)
            result.update(self.get_dataset_split(data_transformer, dataset_type, combined_dataset))
        return result
    
class IndexSplitAlgo(SplitAlgoBase):
    
    CONFIG_FORMAT = {
        "indices":{
            "train": ConfigComponent(dtypes="DICT", required=True,
                                     description="a dictionary containing that maps the sample to the corresponding "
                                                 "selection indices for the train dataset in the form "
                                                 "{<sample_name>: <array of index values>}"),
            "validation": ConfigComponent(dtypes=["DICT", "NONE"], default=None,
                                          description="a dictionary containing that maps the sample to the corresponding "
                                                      "selection indices for the validation dataset in the form "
                                                      "{<sample_name>: <array of index values>}; if none, no validation "
                                                      "dataset will be loaded"),
            "test": ConfigComponent(dtypes=["DICT", "NONE"], default=None,
                                    description="a dictionary containing that maps the sample to the corresponding "
                                                "selection indices for the test dataset in the form "
                                                "{<sample_name>: <array of index values>}; if none, no test "
                                                "dataset will be loaded"),
        },
        "index_variable": ConfigComponent(dtypes="STR", required=True,
                                          description="the variable used for indexing")
    }

    def get_extra_variables(self):
        index_variable = self.config["index_variable"]
        return [index_variable]
    
    def get_split_indices(self):
        split_indices = {}
        split_indices["train"] = self.config["indices"]["train"]
        if "validation" in self.config["indices"]:
            split_indices["val"] = self.config["indices"]["validation"]
        if "test" in self.config["indices"]:
            split_indices["test"] = self.config["indices"]["test"]
        return split_indices
    
    def train_val_test_split(self, analysis_datasets:Dict, data_transformer:DataTransformer):
        
        index_variable = self.config["index_variable"]
        self.stdout.info("INFO: Data will be split using custom event indices.")
        self.stdout.info(f"       Index variable: {index_variable}")
        
        analysis_dataset = analysis_datasets[None]
        concat_kwargs = {
            "axis": 1,
            "sort": False,
            "copy": False
        }
        train_variables = None
        sample_dfs = {}
        for sample, sample_dataset in analysis_dataset.items():
            x, y, weight = sample_dataset["x"], sample_dataset["y"], sample_dataset["weight"]
            if weight is None:
                sample_dfs[sample] = pd.concat([x, y], **concat_kwargs)
            else:
                sample_dfs[sample] = pd.concat([x, y, weight], **concat_kwargs)
            sample_dfs[sample].set_index(index_variable, inplace=True)
        else:
            train_variables = list(x.columns)
        split_indices = self.get_split_indices()
        result = {}
        for dataset_type, indices in split_indices.items():
            dataset_i = {}
            unique_indices = {}
            for sample, sample_df in sample_dfs.items():
                sample_indices = indices.get(sample, None)
                if sample_indices is None:
                    raise RuntimeError(f'missing event indices for the sample "{sample}" when selecting events '
                                       f'for the "{dataset_type}" dataset')
                # need to make sure the event indices are unique to avoid further duplication
                sample_indices = np.unique(sample_indices)
                unique_indices[sample] = sample_indices
                missing_indices = np.setdiff1d(sample_indices, sample_df.index.values)
                if len(missing_indices) > 0:
                    raise RuntimeError(f'failed to select events from the given indices for the '
                                       f'sample "{sample}" for the "{dataset_type}" dataset')
                df_selected = sample_df.loc[sample_indices].reset_index()
                # make sure all data are copied so any subsequent transformation only act on the copy
                dataset_i[sample] = {
                    "x": df_selected[train_variables],
                    "y": df_selected[["true_label"]]
                }
                if "weight" in df_selected.columns:
                    dataset_i[sample]["weight"] = df_selected[["weight"]]
            result.update(self.get_dataset_split(data_transformer, dataset_type, dataset_i, unique_indices))
        return result
    
class ManualSplitAlgo(SplitAlgoBase):
    
    CONFIG_FORMAT = {
        "datasets": ConfigComponent(dtypes=["LIST[STR]", "NONE"], default=None,
                                    description="dataset(s) that should be used in the manual splitting; if none, the "
                                                "full dataset will be used"),
        "split_func": ConfigComponent(dtypes="ANY", required=True,
                                      description="A callable with signature (x, y, weight, **kwargs) -> "
                                                  "{<dataset_name>: (x, y, weight)}"),
        "extra_variables": ConfigComponent(dtypes="LIST[STR]", default=[],
                                           description="extra variables needed for the manual split method")     
    }
    
    def get_analysis_dataset_names(self) -> List[str]:
        dataset_config = self.config.get("datasets", None)
        if dataset_config is not None:
            dataset_names = list(dataset_config)
        else:
            dataset_names = None
        return None
    
    def get_extra_variables(self):
        return self.config.get("extra_variables", [])
    
    def train_val_test_split(self, analysis_datasets:Dict, data_transformer:DataTransformer):
        
        self.stdout.info("INFO: Data will be split using a user-defined function.")
        
        split_func = self.config["split_func"]
        extra_variables = self.config.get("extra_variables", [])
        analysis_dataset = self.combine_analysis_datasets(list(analysis_datasets.values()))
        splitted_dataset = {}
        for sample, sample_dataset in analysis_dataset.items():
            x, y, weight = sample_dataset["x"], sample_dataset["y"], sample_dataset["weight"]
            splitted_sample_dataset = split_func(x, y, weight, sample=sample)
            for dataset_type in splitted_sample_dataset:
                if dataset_type not in splitted_dataset:
                    splitted_dataset[dataset_type] = {}
                splitted_dataset[dataset_type][sample] = splitted_sample_dataset[dataset_type]
        result = {}
        for dataset_type, dataset_i in splitted_dataset.items():
            result.update(self.get_dataset_split(data_transformer, dataset_type, dataset_i))
        return result

class DatasetSplitMethod(DescriptiveEnum):
    KFOLD  = (0, "select event using the k-fold method", KFoldSplitAlgo)
    CUSTOM = (1, "select event using event number modularity", CustomSplitAlgo)
    INDEX  = (2, "select event using from a given set of indices", IndexSplitAlgo)
    MANUAL = (3, "select event using a custom function", ManualSplitAlgo)
    
    def __new__(cls, value:int, description:str, split_algo:SplitAlgoBase):
        obj = object.__new__(cls)
        obj._value_ = value
        obj.description = description
        obj.split_algo = split_algo
        return obj
    
class DataLoader(AnalysisBase):
    
    REQUIRED_CONFIG_COMPONENTS = ["paths:arrays", "samples:*", "kinematic_regions:*",
                                  "variables:*", "training:*", "channels:*",
                                  "names:weight", "names:event_number"]
    
    DEFAULT_DATASET_SPLIT_SETUP = {
        "method": "kfold",
        "options": {
            "datasets": None,
            "index_variable": None,
            "n_splits": 5,
            "random_state": -1
        }
    }
    
    def __init__(self, analysis_config:Union[Dict, str],
                 array_dir:Optional[str]=None,
                 model_dir:Optional[str]=None,
                 verbosity:Optional[Union[int, str]]="INFO",
                 **kwargs):
        super().__init__(analysis_config=analysis_config,
                         array_dir=array_dir,
                         model_dir=model_dir,
                         verbosity=verbosity,
                         **kwargs)
        self.set_data_transformer()
        self.set_split_algo()
        self.data_format = self.get_analysis_data_format()
        
    def set_data_transformer(self, data_transformer:Optional[DataTransformer]=None):
        if data_transformer is None:
            train_config = self.config["training"]
            sample_transformation  = train_config.get("sample_transformation", {})
            dataset_transformation = train_config.get("dataset_transformation", {})
            data_transformer = DataTransformer(sample_transformation=sample_transformation,
                                               dataset_transformation=dataset_transformation)
        self.data_transformer = data_transformer
        
    def set_split_algo(self, split_algo:Optional[SplitAlgoBase]=None):
        if split_algo is None:
            try:
                split_method = self.config["training"]["datasets"]["split_method"]
                split_options = self.config["training"]["datasets"]["split_options"]
            except:
                split_method = self.DEFAULT_DATASET_SPLIT_SETUP["method"]
                split_options = self.DEFAULT_DATASET_SPLIT_SETUP["options"]
            split_method = DatasetSplitMethod.parse(split_method)
            split_algo = split_method.split_algo(split_options,
                                                 verbosity=self.stdout.verbosity)
        self.split_algo = split_algo
        
    def set_random_state(self, random_state:Optional[int]=-1):
        self.data_transformer.set_random_state(random_state)
        self.split_algo.set_random_state(random_state)

    def _format_selection(self, samples:List[str],
                          old_selection:Optional[Union[Dict, str]]=None,
                          new_selection:Optional[Union[Dict, str]]=None):
        if old_selection is None:
            selection = {sample: None for sample in samples}
        elif isinstance(old_selection, str):
            selection = {sample: old_selection for sample in samples}
        elif isinstance(old_selection, dict):
            selection = {sample: old_selection.get(sample, None) for sample in samples}
        else:
            raise ValueError("invalid selection format")
        if new_selection is None:
            return selection
        if isinstance(new_selection, str):
            for sample in samples:
                if selection[sample] is None:
                    selection[sample] = new_selection
                else:
                    selection[sample] = f"({selection['sample']}) and ({new_selection})"                    
        elif isinstance(new_selection, dict):
            for sample, sample_new_selection in new_selection.items():
                if sample not in samples:
                    continue
                if selection[sample] is None:
                    selection[sample] = sample_new_selection
                else:
                    selection[sample] = f"({selection['sample']}) and ({sample_new_selection})"
        return selection

    @semistaticmethod
    def _read_data_from_file(self, filename:str, key:Optional[str]=None,
                             columns:Optional[List[str]]=None,
                             selection:Optional[str]=None) -> pd.DataFrame:
        """
            Load array data from a file (csv or h5).
            
            Arguments:
                filename: string
                    Path to the input data (only csv or h5 formats are allowed).
                selection: (optional) str
                    Selection applied to the input data (via pandas.DataFrame.query)
        """        
        extension = os.path.splitext(filename)[-1]
        if extension not in [".csv", ".h5"]:
            raise RuntimeError(f'unsupported data format "{extension.strip(".")}"; allowed '
                               'formats are: csv, h5')
        if extension == ".csv":
            df = pd.read_csv(filename, usecols=columns)
        else:
            df = pd.read_hdf(filename, key=key, columns=columns)
        if selection not in [None, "1", 1]:
            df = df.query(selection).reset_index(drop=True)
        return df

    def get_sample_df(self, sample:str, key:Optional[str]=None,
                      columns:Optional[List[str]]=None,
                      selection:Optional[str]=None,
                      kinematic_region:Optional[str]=None)->pd.DataFrame:
        if kinematic_region is not None:
            sample_path = self.get_file("train_sample",
                                        subdirectory=kinematic_region,
                                        sample=sample,
                                        fmt=self.data_format)
        else:
            sample_path = self.get_file("train_sample",
                                        sample=sample,
                                        fmt=self.data_format)
        self.stdout.info(f'INFO: Loading inputs for the sample "{sample}" from "{sample_path}"')
        df = self._read_data_from_file(sample_path, key=key, columns=columns, selection=selection)
        return df
    
    def get_sample_dfs(self, samples:List[str], key:Optional[str]=None,
                       columns:Optional[List[str]]=None,
                       selection:Optional[Union[Dict, str]]=None,
                       kinematic_region:Optional[str]=None)->pd.DataFrame:
        dfs = {}
        selection = self._format_selection(samples, selection)
        for sample in samples:
            sample_selection = selection.get(sample, None)
            df = self.get_sample_df(sample=sample, key=key,
                                    columns=columns,
                                    selection=sample_selection,
                                    kinematic_region=kinematic_region)
            dfs[sample] = df
        return dfs
    
    def get_analysis_datasets(self, samples:List[str], train_variables:List[str],
                              class_label_map:Dict,
                              dataset_names:Optional[List[str]]=None,
                              key:Optional[str]=None, 
                              weight_variable:Optional[str]=None,
                              selection_variables:Optional[List[str]]=None,
                              selection:Optional[Union[Dict, str]]=None,
                              kinematic_region:Optional[str]=None):
        for sample in samples:
            if sample not in class_label_map:
                raise RuntimeError(f'missing class label definition for the sample "{sample}"')
        columns = list(train_variables)
        if weight_variable is not None:
            columns.append(weight_variable)
        if selection_variables is not None:
            columns.extend(selection_variables)
        sample_dfs_map = {}
        if dataset_names is not None:
            dataset_selections = {}
            additional_columns = set()
            try:
                spec = self.config["training"]["datasets"]["specification"]
            except:
                spec = {}            
            for dataset_name in list(set(dataset_names)):
                if dataset_name not in spec:
                    raise RuntimeError(f'missing specification for the dataset "{dataset_name}" in '
                                       'analysis config')
                if 'selection' not in spec[dataset_name]:
                    raise RuntimeError(f'missing definition for selection of the dataset "{dataset_name}" '
                                       'in analysis config')
                dataset_selections[dataset_name] = spec[dataset_name]['selection']
                if ('variables' in spec[dataset_name]) and (spec[dataset_name]['variables'] is not None):
                    additional_columns |= set(spec[dataset_name]['variables'])
            columns = list(set(columns).union(additional_columns))
            sample_dfs = self.get_sample_dfs(samples, key=key, columns=columns,
                                             selection=selection,
                                             kinematic_region=kinematic_region)

            for dataset_name in dataset_selections:
                sample_dfs_map[dataset_name] = {}
                selection = dataset_selections[dataset_name]
                for sample, sample_df in sample_dfs.items():
                    sample_dfs_map[dataset_name][sample] = sample_df.query(selection).reset_index(drop=True)
        else:
            sample_dfs_map[None] = self.get_sample_dfs(samples, key=key, columns=columns,
                                                       selection=selection,
                                                       kinematic_region=kinematic_region)

        datasets = {}
        for dataset_name, sample_dfs in sample_dfs_map.items():
            datasets[dataset_name] = {}
            for sample, sample_df in sample_dfs.items():
                data_size = len(sample_df)
                class_label = class_label_map[sample]
                datasets[dataset_name][sample] = {
                    "x": sample_df[train_variables],
                    "y": pd.DataFrame(np.full((data_size,), class_label), columns=['true_label'])
                }
                if weight_variable is None:
                    datasets[dataset_name][sample]["weight"] = None
                else:
                    datasets[dataset_name][sample]["weight"] = sample_df[[weight_variable]]
        return datasets

    def load_train_data(self, samples:List[str],
                        train_variables:List[str],
                        class_label_map:Dict,
                        weight_variable:Optional[str]=None,
                        selection_variables:Optional[List[str]]=None,
                        selection:Optional[Union[Dict, str]]=None,
                        kinematic_region:Optional[str]=None,
                        split_algo:Optional[SplitAlgoBase]=None):
        kwargs = {
            "samples"             : samples,
            "class_label_map"     : class_label_map,
            "weight_variable"     : weight_variable,
            "selection_variables" : selection_variables,
            "selection"           : selection,
            "kinematic_region"    : kinematic_region
        }
        if split_algo is None:
            split_algo = self.split_algo
        extra_variables = split_algo.get_extra_variables()
        for variable in extra_variables:
            if variable in train_variables:
                raise RuntimeError(f'the additional variable "{variable}" used in dataset splitting is '
                                    'already part of the train variables, please double check your input')
        dataset_names = split_algo.get_analysis_dataset_names()
        kwargs["train_variables"] = list(train_variables)
        kwargs["train_variables"].extend(extra_variables)
        
        analysis_datasets = self.get_analysis_datasets(dataset_names=dataset_names, **kwargs)
        
        train_data =  split_algo.train_val_test_split(analysis_datasets, self.data_transformer)
        
        return train_data
                                  
    def get_class_label_map(self, class_labels:Dict):
        labels = list(class_labels.keys())
        encodings = get_class_label_encodings(labels)
        class_label_map = {}
        for label, samples in class_labels.items():
            resolved_samples = self.resolve_samples(samples)
            for sample in resolved_samples:
                if sample in class_label_map:
                    raise RuntimeError(f'multiple class label found for the sample "{sample}"')
                class_label_map[sample] = encodings[label]
        return class_label_map                                  
    
    def load_channel_train_data(self, channel:str, scale_factors:Optional[Dict]=None,
                               custom_indices:Optional[Dict[str, Dict[str, np.ndarray]]]=None,
                               index_variable:Optional[str]=None):
        if channel not in self.all_channels:
            raise ValueError(f"unknown channel: {channel}")
            
        channel_config  = self.config["channels"][channel]
        training_config = self.config["training"]
        class_labels = channel_config['class_labels']
        class_label_map = self.get_class_label_map(class_labels)
        
        if scale_factors is None:
            scale_factors = channel_config.get('SF', {})
        self.data_transformer.set_sample_weight_scales(scale_factors)
        
        if custom_indices is not None:
            if index_variable is None:
                index_variable = self.get_event_index_variable()
            split_options = {
                "indices": custom_indices,
                "index_variable": index_variable
            }
            split_algo = IndexSplitAlgo(split_options)
        else:
            split_algo = self.split_algo
            
        kwargs = {
            "samples"                : self.resolve_samples(channel_config['train_samples']),
            "train_variables"        : self.resolve_variables(channel_config['train_variables']),
            "class_label_map"        : class_label_map,
            "weight_variable"        : self.config['names'].get("weight", None),
            "selection"              : channel_config.get('selection', None),
            "selection_variables"    : channel_config.get('selection_variables', None),
            "kinematic_region"       : channel_config.get('kinematic_region', None),
            "split_algo"             : split_algo
        }
        self.stdout.info(f"INFO: Loading data for the channel: {channel}")
        return self.load_train_data(**kwargs)
        
    def print_channel_summary(self, channels:Optional[Union[List[str],str]]=None):
        if channels is None:
            channels = self.all_channels
        elif isinstance(channels, str):
            channels = [channels]
        for channel in channels:
            if channel not in self.all_channels:
                raise ValueError(f"unknwon channel: {channel}")
            selection = self.config['channels'][channel]['selection']
            train_samples = self.config['channels'][channel]['train_samples']
            test_samples = self.config['channels'][channel]['test_samples']
            train_variables = self.config['channels'][channel]['train_variables']
            class_labels = self.config['channels'][channel]['class_labels']
            train_samples = self.resolve_samples(train_samples)
            test_samples = self.resolve_samples(test_samples)
            train_variables = self.resolve_variables(train_variables)
            scale_factors = self.config['channels'][channel]['SF']
            self.stdout.info("=============================== CHANNEL SUMMARY ===============================")
            self.stdout.info("Channel Name: ".rjust(20) + f"{channel}")
            self.stdout.info("Channel Selection: ".rjust(20) + f"{selection}")
            self.stdout.info("*******************************************************************************")
            self.stdout.info("Train Samples: ".rjust(20) + ", ".join(train_samples))
            self.stdout.info("Test Samples: ".rjust(20) + ", ".join(test_samples))
            self.stdout.info("*******************************************************************************")
            self.stdout.info("Class Labels: ".rjust(20))
            for label in class_labels:
                samples = self.resolve_samples(class_labels[label])
                self.stdout.info(f"\t{label}: " + ", ".join(samples))
            self.stdout.info("*******************************************************************************")
            self.stdout.info("Train Variables: ".rjust(20) + ", ".join(train_variables))
            self.stdout.info("*******************************************************************************")
            self.stdout.info("Scale Factors: ".rjust(20) + str(scale_factors))
            self.stdout.info("*******************************************************************************")
            self.stdout.info("")
            
            
def index_modularity_split(index_variable:str, modulo:int,
                           train_remainders:List[int],
                           val_remainders:Optional[List[int]]=None,
                           test_remainders:Optional[List[int]]=None):
    def split_func(x, y, weight=None, **kwargs):
        masks = {}
        masks["train"] = (x[index_variable] % modulo).isin(train_remainders)
        if val_remainders is not None:
            masks["val"] = (x[index_variable] % modulo).isin(val_remainders)
        if test_remainders is not None:
            masks["test"] = (x[index_variable] % modulo).isin(test_remainders)
        result = {}
        for dataset_type, mask in masks.items():
            result[dataset_type] = {
                "x": x[mask],
                "y": y[mask]
            }
            if weight is not None:
                result[dataset_type]["weight"] = weight[mask]
        return result
    return split_func


from typing import Optional

from .rooproc_helper_action import RooProcHelperAction

from quickstats.utils.root_utils import declare_expression

class RooProcDeclare(RooProcHelperAction):
    
    def __init__(self, expression:str, name:Optional[str]=None):
        super().__init__(expression=expression,
                         name=name)
        
    @classmethod
    def parse(cls, main_text:str, block_text:Optional[str]=None):
        return cls(expression=main_text)        
    
    def _execute(self, processor:"quickstats.RooProcessor", **params):
        name = params.get("name", None)
        expression = params['expression']
        declare_expression(expression, name)
        return processor


###############################################################################
### This is a reimplementation of workspaceCombiner library in python
### original author: Hongtao Yang
###############################################################################
import os
import re
import json
from typing import Optional, Union, List, Dict

try:
    from tabulate import tabulate
except:
    tabulate = None
    
import numpy as np

import ROOT

import quickstats
from quickstats import semistaticmethod, AbstractObject, GeneralEnum, Timer
from quickstats.components import ExtendedModel
from quickstats.utils.io import Verbosity
from quickstats.utils.xml_tools import TXMLTree
from quickstats.utils.root_utils import load_macro, get_macro_dir, is_corrupt
from quickstats.utils.common_utils import format_delimiter_enclosed_text
from quickstats.maths.numerics import is_float, pretty_value
from quickstats.components.workspaces import XMLWSBase, WSObjectType, BracketType

class XMLWSCombiner(XMLWSBase):
    
    KEYWORDS = {
        "dummy": "dummy",
        "constrterm_suffix": "_Pdf",
        "globalobs_suffix": "_In",
        "weight_name": "_weight_",
        "comb_cat_name": "combCat",
        "comb_pdf_name": "combPdf",
        "comb_ws_name": "combWS",
        "comb_data_name": "combData",
        "comb_mc_name": "ModelConfig",
        "comb_nuis_name": "nuisanceParameters",
        "comb_glob_name": "globalObservables",
        "comb_obs_name": "observables",
        "rename_save_suffix": "rename_cache",
        "rename_comb_suffix": "combine_cache"
    }
    
    def __init__(self, source:Union[str, Dict],
                 basedir:Optional[str]=None,
                 minimizer_config:Optional[Dict]=None,
                 unlimited_stack=True,
                 verbosity:Optional[Union[int, str]]="INFO") -> None:
        super().__init__(source=source, basedir=basedir,
                         unlimited_stack=unlimited_stack,
                         verbosity=verbosity)
        self.minimizer_config = minimizer_config
        self.load_extension()
        with Timer() as t:
            self.initialize(source=source)
        self.init_time = t.interval
        
    def initialize(self, source:Union[str, Dict]) -> None:
        self.combination_config = {}
        self.channel_config     = {}
        self.asimov_actions     = []
        if isinstance(source, str):
            ext = os.path.splitext(source)[-1]
            if ext == ".xml":
                self.parse_config_xml(source)
            elif ext == ".json":
                self.parse_config_json(source)
            else:
                raise ValueError(f"unsupported file format: {ext}")
        elif isinstance(source, dict):
            self.parse_config_dict(source)
        else:
            raise ValueError(f"invalid input format: {source}")
        self.reset_tmp_ws()
        self.reset_comb_ws()
 
    def parse_config_xml(self, filename:str) -> None:
        root = TXMLTree.load_as_dict(filename)
        combination_config = {
            'output_file'      : self._get_node_attrib(root, "OutputFile", required=True),
            'workspace_name'   : self._get_node_attrib(root, "WorkspaceName", required=False,
                                                       default=self.KEYWORDS["comb_ws_name"]),
            'model_config_name': self._get_node_attrib(root, "ModelConfigName", required=False,
                                                       default=self.KEYWORDS["comb_mc_name"]),
            'data_name'        : self._get_node_attrib(root, "DataName", required=False,
                                                       default=self.KEYWORDS["comb_data_name"]),
            'strict'           : self._get_node_attrib(root, "Strict", required=False, default="false", dtype="bool")
        }
        self.combination_config = combination_config
        # parse child level attributes
        nodes = root['children']
        for node in nodes:
            tag = node['tag']
            if tag == "Channel":
                self.read_channel_node(node)
            elif tag == "POIList":
                self.read_combined_poi_node(node)
            elif tag == "Asimov":
                self.read_asimov_node(node)
            else:
                raise RuntimeError(f"unknown item `{tag}`")
    
    def parse_config_dict(self, source:Dict) -> None:
        raise NotImplementedError("not implemented")
    
    def parse_config_json(self, filename:str) -> None:
        with open(filename, 'r') as f:
            source = json.load(f)
        self.parse_config_dict(source)
        
    def _get_channel_model(self, channel_config:Dict) -> ExtendedModel:
        input_file = channel_config["input_file"]
        ws_name    = channel_config["workspace_name"]
        mc_name    = channel_config["model_config_name"]
        data_name  = channel_config["data_name"]
        model = ExtendedModel(input_file, ws_name=ws_name, mc_name=mc_name,
                              data_name=data_name, verbosity="WARNING")
        return model
    
    def _exception_check(self, msg:str, strict_mode:bool=False) -> None:
        if strict_mode:
            raise RuntimeError(msg)
        else:
            msg = msg[0].capitalize() + msg[1:]
            self.stdout.warning(f"WARNING: {msg}. Skipped.")
    
    def _channel_check(self, channel:str, check_poi_map:bool=False):
        if channel not in self.channel_config:
            raise RuntimeError(f"channel \"{channel}\" not initialized")
        if check_poi_map and ("poi_map" not in self.channel_config[channel]):
            raise RuntimeError(f'missing poi definition for the channel "{channel}"')
            
    def reset_tmp_ws(self):
        workspace_name = self.combination_config["workspace_name"]
        self.tmp_file = None
        self.tmp_ws   = ROOT.RooWorkspace(workspace_name, workspace_name)
        self.tmp_nuis = ROOT.RooArgSet()
        self.tmp_glob = ROOT.RooArgSet()
        
    def load_cache_ws(self, ws_type:str):
        if ws_type == "rename":
            filename = self._get_tmp_ws_path()
        elif ws_type == "combine":
            filename = self._get_comb_ws_path()
        else:
            raise ValueError(f"invalid cache workspace type: {ws_type}")
        self.stdout.info(f'INFO: Reading temporary workspace file from "{filename}"')
        if not os.path.exists(filename):
            raise FileNotFoundError(f'temporary workspace file "{filename}" does not exist')
        file = ROOT.TFile(filename)
        if is_corrupt(file):
            raise RuntimeError(f'temporary workspace file "{filename}" is corrupted')
        workspace_name = self.combination_config["workspace_name"]
        workspace = file.Get(workspace_name)
        if not workspace:
            raise RuntimeError(f'failed to load workspace "{workspace_name}"')
        nuis_set_name = self.KEYWORDS['comb_nuis_name']
        glob_set_name = self.KEYWORDS['comb_glob_name']
        nuis_set = workspace.set(nuis_set_name)
        glob_set = workspace.set(glob_set_name)
        if not nuis_set:
            raise RuntimeError(f'failed to load nuisance parameter set "{nuis_set_name}"')
        if not glob_set:
            raise RuntimeError(f'failed to load global observable set "{glob_set_name}"')
        if ws_type == "rename":
            self.tmp_file = file
            self.tmp_ws = workspace
            self.tmp_nuis = nuis_set.snapshot()
            self.tmp_glob = glob_set.snapshot()
        elif ws_type == "combine":
            self.comb_file = file
            self.comb_ws = workspace
            self.tmp_nuis = nuis_set.snapshot()
            self.tmp_glob = glob_set.snapshot()
            obs_set_name = self.KEYWORDS['comb_obs_name']
            obs_set = workspace.set(obs_set_name)
            if not obs_set:
                raise RuntimeError(f'failed to load observable set "{obs_set_name}"')
            self.comb_obs = obs_set
            
    def reset_comb_ws(self):
        workspace_name = self.combination_config["workspace_name"]
        self.comb_file = None
        self.comb_ws   = ROOT.RooWorkspace(workspace_name, workspace_name)
        cat_name = self.KEYWORDS['comb_cat_name']
        pdf_name = self.KEYWORDS['comb_pdf_name']
        self.comb_cat = ROOT.RooCategory(cat_name, cat_name)
        self.comb_pdf = ROOT.RooSimultaneous(pdf_name, pdf_name, self.comb_cat)
        self.comb_obs = ROOT.RooArgSet()
        self.comb_data_map = {}
        
    def read_combined_poi_node(self, node:Dict) -> None:
        poi_expressions = self._get_node_attrib(node, "Combined", required=True, dtype="str_list")
        poi_definitions = {}
        for poi_expr in poi_expressions:
            poi_name, content = self._extract_bracket_expr(poi_expr, BracketType.SQUARE)
            if not poi_name:
                raise ValueError(f"invalid format for a combined POI: {poi_expr}")
            if poi_name in poi_definitions:
                raise RuntimeError(f"Combined POI \"{poi_name}\" is defined more than once.")
            tokens = content.split("~")
            tokens_float = []
            for token in tokens:
                if not is_float(token):
                    raise ValueError(f"invalid numerical value \"{token}\" from the POI "
                                     f"expression \"{poi_expr}\"")
                tokens_float.append(float(token))
            poi_info = {
                "value": None,
                "min": None,
                "max": None
            }
            # only central value provided
            if len(tokens) == 1:
                poi_info["value"] = tokens_float[0]
            # only range provided
            elif len(tokens) == 2:
                if abs(tokens_float[0] - tokens_float[1]) < self.EPSILON:
                    poi_info["value"] = tokens_float[0]
                else:
                    poi_info["min"] = tokens_float[0]
                    poi_info["max"] = tokens_float[1]
            # central value and range provided
            elif len(tokens) == 3:
                    poi_info["value"] = tokens_float[0]
                    poi_info["min"]   = tokens_float[1]
                    poi_info["max"]   = tokens_float[2]
            else:
                raise ValueError(f"too many/few values to unpack from the POI expression \"{poi_expr}\"")
            poi_definitions[poi_name] = poi_info
        self.combination_config["poi_definition"] = poi_definitions
    
    def read_asimov_node(self, node:Dict) -> None:
        asimov_action = node['attrib']
        self.asimov_actions.append(asimov_action)

    def read_channel_node(self, node:Dict) -> None:
        channel_config = {
            'channel_name'      : self._get_node_attrib(node, "Name", required=True),
            'input_file'        : self._get_node_attrib(node, "InputFile", required=True),
            'workspace_name'    : self._get_node_attrib(node, "WorkspaceName", required=False),
            'model_config_name' : self._get_node_attrib(node, "ModelConfigName", required=False),
            'data_name'         : self._get_node_attrib(node, "DataName", required=False, default="combData"),
            'simplified_import' : self._get_node_attrib(node, "SimplifiedImport", required=False,
                                                        default="false", dtype="bool"),
            'argset_list'       : [],
            'argset_strict'     : False
        }
        channel = channel_config["channel_name"]
        if channel in self.channel_config:
            raise RuntimeError(f"the channel \"{channel}\" is included more than once")
        self.channel_config[channel] = channel_config
        self.channel_config[channel]['rename_map'] = {
            "pdf"  : {},
            "var"  : {}
        }
        self.channel_config[channel]['renamed_constraint'] = {}
        self.stdout.info(f"INFO: Checking XML input for the channel \"{channel}\".")
        # walk through child nodes
        child_nodes = node["children"]
        for child_node in child_nodes:
            tag = child_node['tag']
            if tag == "RenameMap":
                self.read_channel_rename_node(channel, child_node)
            elif tag == "POIList":
                self.read_channel_poi_node(channel, child_node)
            elif tag == "ArgSetList":
                self.read_channel_argset_node(channel, child_node)
            else:
                raise RuntimeError(f"unknown item `{tag}`")
        self.stdout.info(f"INFO: Successfully validated XML input for the channel \"{channel}\".")
    
    def read_channel_rename_node(self, channel:str, node:Dict) -> None:
        ext_file = self._get_node_attrib(node, "InputFile", required=False)
        if ext_file is not None:
            node = TXMLTree.load_as_dict(ext_file)
        child_nodes = node['children']
        for child_node in child_nodes:
            tag = child_node['tag']
            if tag == "Syst":
                self.read_rename_syst_node(channel, child_node)
            else:
                raise RuntimeError(f"unknown channel item `{tag}`")
                
    def read_channel_argset_node(self, channel:str, node:Dict) -> None:
        argset_list = self._get_node_attrib(node, "Input", required=False,
                                            default=[], dtype="str_list")
        argset_strict = self._get_node_attrib(node, "Strict", required=False,
                                              default="True", dtype="bool")
        self.channel_config[channel]['argset_list'] = argset_list
        self.channel_config[channel]['argset_strict'] = argset_strict
                
    def validate_channel_input(self, channel:str):
        self._channel_check(channel)
        self.stdout.info(f"INFO: Checking input workspace for the channel \"{channel}\".")
        channel_config = self.channel_config[channel]
        model = self._get_channel_model(channel_config)
        strict_mode = self.combination_config["strict"]
        ws_filename = model.fname
        renamed_pdfs = list(channel_config["rename_map"]["pdf"])
        renamed_vars = list(channel_config["rename_map"]["var"])
        renamed_pois = list(channel_config["poi_map"].values())
        renamed_constraint = channel_config["renamed_constraint"]
        missing_obj_msg = f'{{obj_type}} "{{obj_name}}" does not exist in the workspace "{ws_filename}"'
        for pdf_name in renamed_pdfs:
            nuis_name = renamed_constraint[pdf_name]["nuis"]
            glob_name = renamed_constraint[pdf_name]["glob"]
            renamed_vars.remove(nuis_name)
            renamed_vars.remove(glob_name)            
            pdf = model.workspace.pdf(pdf_name)
            if not pdf:
                msg = missing_obj_msg.format(obj_type="Constraint PDF", obj_name=pdf_name)
                self._exception_check(msg, strict_mode)
                continue
            if not isinstance(pdf, ROOT.RooGaussian):
                msg = f"constraint PDF \"{pdf_name}\" that is not an instance of RooGaussian is not supported"
                self._exception_check(msg, strict_mode)
                continue
            nuis = model.workspace.var(nuis_name)
            if not nuis:
                msg = missing_obj_msg.format(obj_type="Nuisance parameter", obj_name=nuis_name)
                raise RuntimeError(msg)
            glob = model.workspace.var(glob_name)
            if not glob:
                msg = missing_obj_msg.format(obj_type="Global observable", obj_name=glob_name)
                raise RuntimeError(msg)
            nuis_check, glob_check, sigma_check, glob_value = False, False, False, None
            # this part can be optimized after ROOT 6.26
            server_iter = pdf.serverIterator()
            while True:
                server = server_iter.Next()
                if not server:
                    break
                server_name = server.GetName()
                if server_name == nuis_name:
                    nuis_check = True
                elif server_name == glob_name:
                    glob_check = True
                    glob_value = server.getVal()
                elif abs(server.getVal() - 1) < self.EPSILON:
                    sigma_check = True
            if (not sigma_check) and (glob_value != 1.):
                raise RuntimeError(f"sigma of constraint pdf \"{pdf_name}\" in the workspace \"{ws_filename}\" is not unity")
        for var_name in renamed_vars:
            if not model.workspace.var(var_name):
                msg = missing_obj_msg.format(obj_type="variable", obj_name=var_name)
                self._exception_check(msg, strict_mode)
                continue
        for poi_name in renamed_pois:
            if poi_name == self.KEYWORDS['dummy']:
                continue
            # check poi exist in workspace
            if not model.workspace.var(poi_name):
                raise RuntimeError(f"poi \"{poi_name}\" does not exist in the channel \"{channel}\"")
        # fix unspecified names
        channel_config['workspace_name'] = model.workspace.GetName()
        channel_config['model_config_name'] = model.model_config.GetName()
        channel_config['data_name'] = model.data.GetName()
        
        argset_list = channel_config["argset_list"]
        for argset in argset_list:
            argset_obj = model.workspace.obj(argset)
            if not argset_obj:
                if channel_config["argset_strict"]:
                    raise RuntimeError(f'Argument set "{argset}" does not exist in the channel "{channel}"')
                else:
                    self.stdout.warning(f'WARNING: Argument set "{argset}" does not exist in the channel "{channel}". Skipped.')
                    continue
            if not isinstance(argset_obj, ROOT.RooArgSet):
                raise RuntimeError(f'Argument set "{argset}" from the channel "{channel}" is an instance of RooArgSet object')
        self.stdout.info(f"INFO: Successfully validated input workspace for the channel \"{channel}\".")
        
    def read_rename_syst_node(self, channel:str, node:Dict) -> None:
        self._channel_check(channel)
        channel_config = self.channel_config[channel]
        old_name = self._get_node_attrib(node, "OldName", required=True)
        new_name = self._get_node_attrib(node, "NewName", required=True)
        if not old_name:
            raise ValueError(f"missing attribute \"OldName\" from the Syst node (channel = {channel}):\n"
                             f"{node}")
        if not new_name:
            raise ValueError(f"missing attribute \"NewName\" from the Syst node (channel = {channel}):\n"
                             f"{node}")
        renamed_constraint = channel_config["renamed_constraint"]
        pdf_map  = channel_config["rename_map"]["pdf"]
        var_map  = channel_config["rename_map"]["var"]
        _, object_type = self._get_object_name_and_type_from_expr(new_name)
        if object_type != WSObjectType.DEFINED:
            raise ValueError(f"the attribute \"NewName\" ({new_name}) from the Syst node should contain "
                             "only the nuisance parameter name")
        new_names = list(var_map.values())
        if new_name in new_names:
            raise ValueError(f"New nuisance parameter \"{new_name}\" is duplicated in the channel \"{channel}\"")
        _, object_type = self._get_object_name_and_type_from_expr(old_name)
        if object_type == WSObjectType.CONSTRAINT:
            pdf_name, content = self._extract_bracket_expr(old_name, BracketType.ROUND)
            components = content.split(",")
            if len(components) != 2:
                raise ValueError(f"invalid format for constraint pdf \"{old_name}\" in channel \"{channel}\"")
            nuis_name = components[0]
            glob_name = components[1]
            if nuis_name in var_map:
                raise ValueError(f"Old nuisance parameter \"{new_name}\" is duplicated in the channel \"{channel}\"")
            var_map[nuis_name] = new_name
            if glob_name in var_map:
                raise ValueError(f"Old global observable \"{glob_name}\" is duplicated in the channel \"{channel}\"")
            var_map[glob_name] = new_name + self.KEYWORDS["globalobs_suffix"]
            if pdf_name in pdf_map:
                raise ValueError(f"Old PDF \"{pdf_name}\" is duplicated in the channel \"{channel}\"")
            pdf_map[pdf_name] = new_name + self.KEYWORDS["constrterm_suffix"]
            renamed_constraint[pdf_name] = {"nuis": nuis_name, "glob": glob_name}
        elif WSObjectType.DEFINED:
            if old_name in var_map:
                raise ValueError(f"Old nuisance parameter \"{old_name}\" is duplicated in the channel \"{channel}\"")
            var_map[old_name] = new_name
                
    def read_channel_poi_node(self, channel:str, node:Dict) -> None:
        self._channel_check(channel)
        channel_config = self.channel_config[channel]
        pois = self._get_node_attrib(node, "Input", required=True, dtype="str_list")
        if "poi_definition" not in self.combination_config:
            raise RuntimeError("Combined POIs not initialized")
        comb_pois = list(self.combination_config["poi_definition"])
        if len(pois) > len(comb_pois):
            raise RuntimeError(f"channel \"{channel}\" has more POIs than the combined model")
        poi_check = {}
        for poi in pois:
            if poi == self.KEYWORDS['dummy']:
                continue
            # check for duplication
            if poi in poi_check:
                raise RuntimeError(f"POI \"{poi}\" for the channel \"{channel}\" is defined more than once.")
            poi_check[poi] = None
        dummy_pois = [self.KEYWORDS['dummy']] * (len(comb_pois) - len(pois))
        pois += dummy_pois
        poi_map = dict(zip(comb_pois, pois))
        channel_config["poi_map"] = poi_map
        
    def _append_channel_suffix_to_args(self, channel:str, args:ROOT.RooArgSet, arg_type:str,
                                       ignore_list:Optional[List[str]]=None) -> None:
        if ignore_list is None:
            ignore_list = []
        for arg in args:
            arg_name = arg.GetName()
            if arg_name in ignore_list:
                continue
            arg.SetName(f"{arg_name}_{channel}")
            self.stdout.debug(f'DEBUG: Appended channel suffix "{channel}" to the {arg_type} '
                              f'"{arg_name}"')
    
    def _get_tmp_ws_path(self):
        basename = os.path.splitext(self.combination_config["output_file"])[0]
        filename = f"{basename}_{self.KEYWORDS['rename_save_suffix']}.root"
        return filename
    
    def _get_comb_ws_path(self):
        basename = os.path.splitext(self.combination_config["output_file"])[0]
        filename = f"{basename}_{self.KEYWORDS['rename_comb_suffix']}.root"
        return filename    
    
    def _get_weight_var(self):
        weight_var = ROOT.RooRealVar(self.KEYWORDS['weight_name'], "", 1.)
        return weight_var
    
    def _get_obs_and_weight(self, observables:ROOT.RooArgSet):
        weight_var = self._get_weight_var()
        obs_and_weight = ROOT.RooArgSet(observables, weight_var)
        return obs_and_weight, weight_var
    
    def get_channel_summary(self, channel:str, indent:int=4) -> str:
        self._channel_check(channel)
        channel_config = self.channel_config[channel]
        summary_text = ""
        summary_text += " "*indent + f" Input File Name: {channel_config['input_file']}" + "\n"
        summary_text += " "*indent + f"  Workspace Name: {channel_config['workspace_name']}" + "\n"
        summary_text += " "*indent + f"ModelConfig Name: {channel_config['model_config_name']}" + "\n"
        summary_text += " "*indent + f"       Data Name: {channel_config['data_name']}" + "\n"
        return summary_text

    def get_poi_table(self) -> str:
        if "poi_definition" not in self.combination_config:
            raise RuntimeError("missing poi definition for the combined channel")
        poi_maps = {}
        poi_maps["Combined"] = list(self.combination_config["poi_definition"])
        for channel in self.channel_config:
            self._channel_check(channel, check_poi_map=True)
            channel_config = self.channel_config[channel]
            channel_poi_map = channel_config['poi_map']
            poi_maps[channel] = []
            for poi in poi_maps["Combined"]:
                if poi not in channel_poi_map:
                    raise RuntimeError(f'missing mapping for the poi "{poi}" from the combined channel '
                                       f'to the channel "{channel}"')
                if poi == self.KEYWORDS['dummy']:
                    poi_maps[channel].append("-")
                else:
                    poi_maps[channel].append(channel_poi_map[poi])
        import pandas as pd
        df = pd.DataFrame(poi_maps)
        if tabulate is None:
            poi_table = df.to_string(index=False)
        else:
            poi_table = tabulate(df, headers='keys', tablefmt='psql', showindex=False)
        return poi_table
    
    def get_summary_text(self, indent:int=4) -> None:
        summary_text = ""
        n_channel = len(self.channel_config)
        title = format_delimiter_enclosed_text(f"Input summary ({n_channel} channels)",
                                               delimiter="-", indent_str="")
        summary_text += title
        for channel in self.channel_config:
            title = format_delimiter_enclosed_text(f"Channel: {channel}",
                                                   delimiter="+", indent_str="")
            summary_text += title
            channel_summary_text = self.get_channel_summary(channel, indent=indent)
            summary_text += channel_summary_text
        title = format_delimiter_enclosed_text(f"POI map", delimiter="-", indent_str="")
        summary_text += title
        poi_table = self.get_poi_table()
        summary_text += poi_table
        return summary_text
    
    def print_input_summary(self) -> None:
        summary_text = self.get_summary_text()
        self.stdout.info(summary_text)
        
    def rename_channel(self, channel:str) -> None:
        self._channel_check(channel, check_poi_map=True)
        channel_config = self.channel_config[channel]
        simplified_import = channel_config["simplified_import"]
        self.stdout.info(f'INFO: Renaming objects for the channel "{channel}"')
        # do not reuse model
        model = self._get_channel_model(channel_config)
        workspace = model.workspace
        variables = workspace.allVars()
        pdfs = workspace.allPdfs()
        functions = workspace.allFunctions()
        ignore_list = []
        # let global observables fixed, and nuisances parameters float
        ROOT.RooStats.SetAllConstant(model.nuisance_parameters, False)
        ROOT.RooStats.SetAllConstant(model.global_observables, True)
        rename_map = channel_config["rename_map"]
        # rename pdf
        for old_name, new_name in rename_map["pdf"].items():
            old_pdf = workspace.pdf(old_name)
            if not simplified_import:
                pdfs.remove(old_pdf)
            if old_name == new_name:
                self.stdout.debug(f'DEBUG: Found redundant renaming of the pdf "{old_name}". Skipped.')
                continue
            old_pdf.SetName(new_name)
            self.stdout.debug(f'DEBUG: Renamed pdf from "{old_name}" to "{new_name}"')
        # rename variable
        for old_name, new_name in rename_map["var"].items():
            old_var = workspace.var(old_name)
            if not simplified_import:
                variables.remove(old_var)
            else:
                ignore_list.append(new_name)
            if old_name == new_name:
                self.stdout.debug(f'DEBUG: Found redundant renaming of the variable "{old_name}". Skipped.')
                continue
            old_var.SetName(new_name)
            self.stdout.debug(f'DEBUG: Renamed variable from "{old_name}" to "{new_name}"')
        # rename pois
        poi_map = channel_config['poi_map']
        for comb_poi_name, channel_poi_name in poi_map.items():
            if channel_poi_name == self.KEYWORDS['dummy']:
                continue
            poi = workspace.var(channel_poi_name)
            if not simplified_import:
                variables.remove(poi)
            else:
                ignore_list.append(comb_poi_name)
            if comb_poi_name == channel_poi_name:
                self.stdout.debug(f'DEBUG: Found redundant renaming of the poi "{channel_poi_name}". Skipped.')
                continue
            poi.SetName(comb_poi_name)
            self.stdout.debug(f'DEBUG: Renamed poi from "{channel_poi_name}" to "{comb_poi_name}"')
        if simplified_import:
            self.stdout.info(f'INFO: Simplified import requested for the channel "{channel}". '
                             'Only unspecified constraint PDFs, nuisance parameters, '
                             'global observables and POIs will be renamed.')
            global_observables = model.global_observables
            nuisance_parameters = model.nuisance_parameters
            pois = model.pois
            constr_pdfs = ROOT.RooArgSet()
            # get the relevant constraint pdfs
            for glob in global_observables:
                glob_name = glob.GetName()
                if glob_name in ignore_list:
                    continue
                client_iter = glob.clientIterator()
                constr_pdf = client_iter.Next()
                if (not constr_pdf) or (not isinstance(constr_pdf, ROOT.RooAbsPdf)):
                    raise RuntimeError(f'failed to find the constraint pdf for the global observable "{glob_name}"')
                constraint_pdfs.add(constr_pdf)
                if client_iter.Next():
                    raise RuntimeError(f'global observable "{glob_name}" in channel "{channel}" '
                                       'has more than one constraint PDF')
            # add channel suffix to global observables
            self._append_channel_suffix_to_args(channel, global_observables,
                                                "global observable", ignore_list)
            # add channel suffix to constraint pdfs
            self._append_channel_suffix_to_args(channel, constr_pdfs,
                                                "constraint PDF")
            # add channel suffix to nuisance_parameters
            self._append_channel_suffix_to_args(channel, nuisance_parameters,
                                                "nuisance parameter", ignore_list)
            # add channel suffix to pois
            self._append_channel_suffix_to_args(channel, pois,
                                                "POI", ignore_list)
        else:
            # remove observables
            observables = ROOT.RooArgSet(model.observables)
            variables.remove(observables)
            
            # add channel suffix to functions
            self._append_channel_suffix_to_args(channel, functions, "function")
            self.stdout.debug(f'DEBUG: All functions in channel "{channel}" have been renamed')
            self._append_channel_suffix_to_args(channel, pdfs, "PDF")
            self.stdout.debug(f'DEBUG: All PDFs in channel "{channel}" have been renamed')
            self._append_channel_suffix_to_args(channel, variables, "variable")
            self.stdout.debug(f'DEBUG: All variables in channel "{channel}" have been renamed')
        
        # rename key objects
        pdf  = model.pdf
        data = model.data
        if not isinstance(pdf, ROOT.RooSimultaneous):
            raise RuntimeError(f'main pdf of the channel "{channel}" does not belong to the RooSimultaneous class')
        cat = pdf.indexCat()
        old_cat_name = cat.GetName()
        new_cat_name = f"{self.KEYWORDS['comb_cat_name']}_{channel}"
        pdf.SetName(f"{self.KEYWORDS['comb_pdf_name']}_{channel}")
        data.SetName(f"{self.combination_config['data_name']}_{channel}")
        
        self.import_object(self.tmp_ws, pdf,
                           ROOT.RooFit.RenameVariable(old_cat_name, new_cat_name),
                           ROOT.RooFit.RecycleConflictNodes(), silent=True)
        self.import_object(self.tmp_ws, data,
                           ROOT.RooFit.RenameVariable(old_cat_name, new_cat_name),
                           silent=False)
        self.update_channel_argset(channel, workspace, self.tmp_ws)
        self.tmp_nuis.add(model.nuisance_parameters.snapshot(), True)
        self.tmp_glob.add(model.global_observables.snapshot(), True)
        
    def update_channel_argset(self, channel:str, orig_ws:ROOT.RooWorkspace, new_ws:ROOT.RooWorkspace):
        channel_config = self.channel_config[channel]
        argset_map = {}
        for argset_name in channel_config["argset_list"]:
            argset_old = orig_ws.obj(argset_name)
            assert isinstance(argset_old, ROOT.RooArgSet)
            components = []
            for arg in argset_old:
                names = [arg.GetName(), f"{arg.GetName()}_{channel}"]
                for name in names:
                    obj = new_ws.obj(name)
                    if obj:
                        components.append(name)
                        break
                else:
                    raise RuntimeError(f'unable to locate the variable "{arg.GetName()}" in the renamed workspace '
                                       f'for the channel {channel}')
            new_argset_name = f"{argset_name}_{channel}"
            if (new_argset_name in self.argsets) or (new_argset_name in argset_map):
                raise RuntimeError(f'duplicated ArgSet object "{new_argset_name}"')
            argset_map[new_argset_name] = components
        self.argsets.update(argset_map)
    
    def rename(self, save_rename_ws:bool=True) -> None:
        title = format_delimiter_enclosed_text(f"Rename PDFs, functions and variables",
                                               delimiter="-", indent_str="")
        self.stdout.info(title)
        self.reset_tmp_ws()
        for channel in self.channel_config:
            self.rename_channel(channel)
        if save_rename_ws:
            self.tmp_ws.defineSet(self.KEYWORDS['comb_nuis_name'], self.tmp_nuis, True);
            self.tmp_ws.defineSet(self.KEYWORDS['comb_glob_name'], self.tmp_glob, True);
            filename = self._get_tmp_ws_path()
            self.stdout.info(f'INFO: Saving temporary workspace to "{filename}"')
            self.tmp_ws.writeToFile(filename, True)
            
    def combine_channel(self, channel:str):
        title = format_delimiter_enclosed_text(f"Channel {channel}",
                                               delimiter="+", indent_str="")
        self.stdout.info(title)
        channel_pdf_name = f"{self.KEYWORDS['comb_pdf_name']}_{channel}"
        channel_data_name = f"{self.KEYWORDS['comb_data_name']}_{channel}"
        channel_cat_name = f"{self.KEYWORDS['comb_cat_name']}_{channel}"
        channel_pdf = self.tmp_ws.pdf(channel_pdf_name)
        if not channel_pdf:
            raise RuntimeError(f'missing pdf "{channel_pdf_name}" in the channel "{channel}"')
        channel_cat = channel_pdf.indexCat()
        channel_data = self.tmp_ws.data(channel_data_name)
        if not channel_data:
            raise RuntimeError(f'missing data "{channel_data_name}" in the channel "{channel}"')
        # split the original dataset
        channel_data_list = channel_data.split(channel_cat, True)
        
        n_cat = channel_cat.size()
        for i in range(n_cat):
            channel_cat.setBin(i)
            cat_name_i = channel_cat.getLabel()
            new_cat_name_i = f"{channel_cat_name}_{cat_name_i}"
            self.stdout.info(f"Category {i+1}: {cat_name_i} --> {new_cat_name_i}")
            pdf_i = channel_pdf.getPdf(cat_name_i)
            data_i = channel_data_list.FindObject(cat_name_i)
            observables_i = pdf_i.getObservables(data_i)
            obs_and_weight, weight_var = self._get_obs_and_weight(observables_i)
            # create new dataset
            new_data_i = ROOT.RooDataSet(f"{new_cat_name_i}_data", f"{new_cat_name_i}_data",
                                         obs_and_weight, ROOT.RooFit.WeightVar(self.KEYWORDS['weight_name']))
            for j in range(data_i.numEntries()):
                observables_i.__assign__(data_i.get(j))
                weight_val = data_i.weight()
                new_data_i.add(obs_and_weight, weight_val)
            self.comb_cat.defineType(new_cat_name_i)
            self.comb_pdf.addPdf(pdf_i, new_cat_name_i)
            self.comb_data_map[new_cat_name_i] = new_data_i
            self.comb_obs.add(observables_i.snapshot(), True)
            
    def combine(self, load_rename_ws:bool=True, save_combine_ws:bool=True) -> None:
        title = format_delimiter_enclosed_text(f"Create Combined Pdf and Dataset",
                                               delimiter="-", indent_str="")
        self.stdout.info(title)
        if load_rename_ws:
            self.load_cache_ws("rename")
        
        self.reset_comb_ws()
        for channel in self.channel_config:
            self.combine_channel(channel)
        
        pdf_name  = self.KEYWORDS['comb_pdf_name']
        data_name = self.combination_config["data_name"]
        self.stdout.info(f'INFO: Generating combined pdf "{pdf_name}"')
        self.comb_ws.Import(self.comb_pdf)
        self.stdout.info(f'INFO: Generating combined dataset "{data_name}"')
        self.comb_obs.add(self.comb_cat)
        obs_and_weight, weight_var = self._get_obs_and_weight(self.comb_obs)
        dataset_map = ExtendedModel.get_dataset_map(self.comb_data_map)
        comb_data = ROOT.RooDataSet(data_name, data_name, obs_and_weight,
                                    ROOT.RooFit.Index(self.comb_cat),
                                    ROOT.RooFit.Import(dataset_map),
                                    ROOT.RooFit.WeightVar(weight_var))
        self.comb_ws.Import(comb_data)
        self.stdout.info(f'INFO: Defining variable set "{self.KEYWORDS["comb_nuis_name"]}"')
        self.comb_ws.defineSet(self.KEYWORDS["comb_nuis_name"], self.tmp_nuis, True)
        self.stdout.info(f'INFO: Defining variable set "{self.KEYWORDS["comb_glob_name"]}"')
        self.comb_ws.defineSet(self.KEYWORDS["comb_glob_name"], self.tmp_glob, True)
        self.stdout.info(f'INFO: Defining variable set "{self.KEYWORDS["comb_obs_name"]}"')
        self.comb_ws.defineSet(self.KEYWORDS["comb_obs_name"], self.comb_obs, True)
        self.define_argsets(self.comb_ws)    
        if save_combine_ws:
            filename = self._get_comb_ws_path()
            self.stdout.info(f'INFO: Saving temporary workspace to "{filename}"')
            self.comb_ws.writeToFile(filename, True)
            
    def define_argsets(self, ws:ROOT.RooWorkspace):
        for argset_name, arg_names in self.argsets.items():
            argset = ROOT.RooArgSet()
            for arg_name in arg_names:
                arg = ws.obj(arg_name)
                if not arg:
                    raise RuntimeError(f'failed to locate object "{arg_name}" from the combined workspace when '
                                       f'defining the ArgSet "{argset_name}"')
                argset.add(arg)
            self.stdout.info(f'INFO: Defining argument set "{argset_name}"')
            ws.Import(argset, argset_name, True)
        
    def get_combined_model_config(self) -> None:
        pois = ROOT.RooArgSet()
        if "poi_definition" not in self.combination_config:
            raise RuntimeError("Combined POIs not initialized")        
        poi_definitions = self.combination_config["poi_definition"]
        for poi_name, poi_info in poi_definitions.items():
            poi = self.comb_ws.var(poi_name)
            if poi:
                poi_min = poi_info["min"]
                poi_max = poi_info["max"]
                poi_value = poi_info["value"]
                if (poi_min is None) and (poi_max is None):
                    poi.setConstant(True)
                else:
                    poi.setRange(poi_min, poi_max)
                    if (poi_value is not None) and ((poi_value > poi_min) and (poi_value < poi_max)):
                        poi.setVal(poi_value)
                pois.add(poi)
            else:
                self.stdout.warning(f'WARNING: The POI "{poi_name}" does not exist in the combined model')
        mc_name = self.combination_config["model_config_name"]
        model_config = ROOT.RooStats.ModelConfig(mc_name, self.comb_ws)
        #model_config.SetWorkspace(self.com_ws)
        pdf_name  = self.KEYWORDS['comb_pdf_name']
        data_name = self.combination_config["data_name"]
        pdf = self.comb_ws.pdf(pdf_name)
        data = self.comb_ws.data(data_name)
        
        model_config.SetPdf(pdf)
        model_config.SetProtoData(data)
        model_config.SetParametersOfInterest(pois)
        model_config.SetNuisanceParameters(self.tmp_nuis)
        model_config.SetGlobalObservables(self.tmp_glob)
        model_config.SetObservables(self.comb_obs)
        
        return model_config
    
    def get_model_config_summary(self, indent:int=0):
        mc_name = self.combination_config["model_config_name"]
        model_config = self.comb_ws.obj(mc_name)
        if not model_config:
            raise RuntimeError("combined model config not defined")
        pdf = model_config.GetPdf()
        if not pdf:
            raise RuntimeError("PDF not defined in ModelConfig")
        cat  = pdf.indexCat()
        nuis = model_config.GetNuisanceParameters()
        if not nuis:
            raise RuntimeError("Nuisance parameters not defined in ModelConfig")        
        glob = model_config.GetGlobalObservables()
        if not glob:
            raise RuntimeError("Global observables not defined in ModelConfig")        
        pois = model_config.GetParametersOfInterest()
        if not pois:
            raise RuntimeError("POIs not defined in ModelConfig")
            
        summary_text = ""
        summary_text += " " * indent + f"Number of categories: {cat.size()}\n"
        summary_text += " " * indent + f"Number of POIs: {pois.getSize()}\n"
        summary_text += " " * indent + f"Number of nuisance parameters: {nuis.getSize()}\n"
        summary_text += " " * indent + f"Number of global observables: {glob.getSize()}\n"
        
        return summary_text
    
    def generate_asimov(self):
        ws = self.comb_ws
        asimov_definitions = self.asimov_actions
        data_name = self.combination_config["data_name"]
        self._generate_asimov(ws, asimov_definitions=asimov_definitions,
                              data_name=data_name,
                              minimizer_config=self.minimizer_config,
                              title_indent_str="")    
        
    def finalize(self, load_combine_ws:bool=True, save_final_ws:bool=True,
                 import_class_code:bool=True) -> None:
        if load_combine_ws:
            self.load_cache_ws("combine")
        
        # create model config
        self.stdout.info(f'INFO: Creating ModelConfig "{self.combination_config["model_config_name"]}"')
        model_config = self.get_combined_model_config()
        self.comb_ws.Import(model_config)
        if import_class_code:
            self.import_class_code(self.comb_ws)
        
        title = format_delimiter_enclosed_text(f"Generate Asimov Dataset",
                                               delimiter="-", indent_str="")
        self.stdout.info(title)
        
        # generate asimov dataset
        self.generate_asimov()
            
        title = format_delimiter_enclosed_text(f"Workspace Summary",
                                               delimiter="-", indent_str="")
        self.stdout.info(title)
        
        mc_summary = self.get_model_config_summary()
        self.stdout.info(mc_summary)
        
        if save_final_ws:
            filename = self.combination_config["output_file"]
            self.stdout.info(f'INFO: Saving final combined workspace to "{filename}"')
            self.comb_ws.writeToFile(filename, True)
            
    def reset_combination(self):
        self.argsets = {}
    
    def create_combined_workspace(self, infiles:Optional[Dict[str, str]]=None,
                                  outfile:Optional[str]=None,
                                  save_rename_ws:bool=False,
                                  save_combine_ws:bool=False,
                                  save_final_ws:bool=True,
                                  import_class_code:bool=True) -> None:
        with Timer() as t:
            # override path to the input channel workspaces
            if infiles is not None:
                for channel in self.channel_config:
                    if channel not in infiles:
                        raise ValueError(f'missing input workspace path for the channel "{channel}"')
                    self.channel_config[channel]['input_file'] = infiles[channel]
            self.reset_combination()
            # override path to the combined workspace
            if outfile is not None:
                self.combination_config['output_file'] = outfile
            for channel in self.channel_config:
                self.validate_channel_input(channel)
            self.print_input_summary()
            self.rename(save_rename_ws=save_rename_ws)
            self.combine(load_rename_ws=save_rename_ws, save_combine_ws=save_combine_ws)
            self.finalize(load_combine_ws=save_combine_ws, save_final_ws=save_final_ws,
                          import_class_code=import_class_code)
        self.stdout.info(f"INFO: Total time taken: {t.interval:.3f}s")


from typing import Optional, Union, Dict, List, Callable
import os
import re
import copy
import glob
import fnmatch
import itertools
from functools import partial

import numpy as np

from quickstats import semistaticmethod, DescriptiveEnum
from quickstats.maths.numerics import pretty_float, str_encode_value, str_decode_value, to_string, to_rounded_float
from quickstats.utils.string_utils import remove_whitespace

signature_regex = {
    'F': r"\d+[.]?\d*",
    'P': r"n?\d+p?\d*",
    'S': r"\w+"
}

signature_parser = {
    'F': pretty_float,
    'P': str_decode_value,
    'S': str
}

signature_regex = {
    'F': r"\d+[.]?\d*",
    'P': r"n?\d+p?\d*",
    'S': r"\w+"
}

signature_parser = {
    'F': pretty_float,
    'P': str_decode_value,
    'S': str
}

class MatchMode(DescriptiveEnum):
    
    Numerical = (0, "Match by value", lambda x, y: to_rounded_float(x) == to_rounded_float(y))
    Wildcard  = (1, "Match by wildcard", lambda x, y: fnmatch.fnmatch(to_string(x), to_string(y)))

    def __new__(cls, value:int, description:str, match_func:Callable):
        obj = object.__new__(cls)
        obj._value_ = value
        obj.description = description
        obj.match_func = match_func
        return obj

class ParamParser:
    
    DEFAULT_FORMAT_STR = r"[\w-]+"
    
    DEFAULT_FILE_EXT = ".root"
    
    @property
    def format_str(self):
        return self._format_str
    
    @format_str.setter
    def format_str(self, val):
        if val is None:
            self._format_str = self.DEFAULT_FORMAT_STR
        else:
            self._format_str = val
        self.fname_regex   = self.get_fname_regex(self._format_str, self.file_ext)
        self.attribute_parser = self.get_attribute_parser(self._format_str)
        
    @property
    def file_ext(self):
        return self._file_ext
    
    @file_ext.setter
    def file_ext(self, val):
        if val is None:
            self._file_ext = self.DEFAULT_FILE_EXT
        else:
            if not val.startswith("."):
                self._file_ext = f".{val}"
            else:
                self._file_ext = val

    def __init__(self, format_str:Optional[str]=None,
                 param_str:Optional[str]=None,
                 file_ext:Optional[str]=None,
                 allow_none:bool=True):
        self.allow_none = allow_none
        self.setup(format_str, param_str, file_ext)
    
    def setup(self, format_str:Optional[str]=None,
              param_str:Optional[str]=None,
              file_ext:Optional[str]=None):
        self.file_ext   = file_ext
        self.format_str = format_str
        self.param_str  = param_str
        self.param_names = self._get_param_names(self.format_str, self.param_str)

    @semistaticmethod
    def _get_param_names(self, format_str:Optional[str]=None,
                        param_str:Optional[str]=None):
        ext_param_names = self._get_format_str_attributes(format_str)
        int_param_names = self._get_param_str_attributes(param_str)
        if set(int_param_names) & set(ext_param_names):
            raise RuntimeError("internal and external parameters are not mutually exclusive")
        return sorted(int_param_names + ext_param_names)
            
    @staticmethod
    def get_signature_map(format_str:str):
        attribute_groups = re.findall(r"<(\w+)\[(\w)\]>", format_str)
        signature_map = {}
        for group in attribute_groups:
            attribute, signature = group[0], group[1]
            signature_map[attribute] = signature
        return signature_map
    
    @staticmethod
    def get_fname_regex(format_str:str, ext:Optional[str]=None):
        if ext is None:
            ext = ParamParser.DEFAULT_FILE_EXT    
        expr = format_str
        signature_map = ParamParser.get_signature_map(format_str)
        for attribute, signature in signature_map.items():
            attribute_expr = signature_regex.get(signature.upper(), None)
            if expr is None:
                raise ValueError(f"unknown signature `{signature}`")
            group_expr = f"(?P<{attribute}>{attribute_expr})"
            expr = expr.replace(f"<{attribute}[{signature}]>", group_expr)
        expr += (ext.replace('.', r'\.') + "$")
        regex = re.compile(expr)
        return regex
   
    @staticmethod
    def sort_param_points(param_points:List, attributes:List):
        key = lambda d: tuple(d['parameters'][attrib] for attrib in attributes)
        return sorted(param_points, key=key)
    
    @semistaticmethod
    def get_attribute_parser(self, format_str:str):
        attribute_parser = {}
        signature_map = self.get_signature_map(format_str)
        for attribute, signature in signature_map.items():
            parser = signature_parser.get(signature, None)
            if parser is None:
                raise ValueError(f"unknown signature `{signature}`")
            attribute_parser[attribute] = parser
        return attribute_parser
    
    @staticmethod
    def _get_param_str_attributes(param_str:Optional[str]=None):
        if param_str is None:
            return []
        attributes = []
        param_expr_list = [s.strip() for s in param_str.split(',')]
        param_expr_list = [s for s in param_expr_list if s]
        for expr in param_expr_list:
            tokens = expr.split('=')
            if len(tokens) not in [1, 2]:
                raise ValueError('invalid expression for parameterisation')
            attribute = tokens[0]
            if attribute not in attributes:
                attributes.append(attribute)
        return attributes
    
    @semistaticmethod
    def _get_format_str_attributes(self, format_str:Optional[str]=None):
        if format_str is None:
            return []
        attribute_parser = self.get_attribute_parser(format_str)
        return list(attribute_parser)

    @staticmethod
    def parse_param_str(param_str:Optional[str]=None):
        if param_str is None:
            return {}
        param_points = []
        idp_param_strs = [s.strip() for s in param_str.split(";")]
        for idp_param_str in idp_param_strs:
            param_values = {}
            param_expr_list = [s.strip() for s in idp_param_str.split(',')]
            param_expr_list = [s for s in param_expr_list if s]
            for expr in param_expr_list:
                tokens = expr.split('=')
                # floating poi
                if len(tokens) == 1:
                    param_name = tokens[0]
                    if param_name in param_values:
                        raise RuntimeError(f"profiled parameter {param_name} appeared more than once in the parameter "
                                           f"expression: {idp_param_str}")
                    param_values[param_name] = [None]
                    continue
                if len(tokens) != 2:
                    raise ValueError('invalid expression for parameterisation')
                param_name = tokens[0]
                values_expr = tokens[1]
                tokens = values_expr.split('_')
                # fixed value
                if len(tokens) == 1:
                    values = [float(tokens[0])]
                # scan across range
                elif len(tokens) == 3:
                    poi_min = float(tokens[0])
                    poi_max = float(tokens[1])
                    poi_step = float(tokens[2])
                    values = np.arange(poi_min, poi_max + poi_step, poi_step)
                else:
                    raise ValueError('invalid expression for parameterisation')
                if param_name not in param_values:
                    param_values[param_name] = np.array([])
                if None in param_values[param_name]:
                    raise RuntimeError(f"the parameter {param_name} is being profiled and non-profiled at the same time "
                                       f"in the parameter expression: {idp_param_str}")
                param_values[param_name] = np.concatenate([param_values[param_name], values])
            param_names = list(param_values.keys())
            # convert to numpy arrays
            combinations = [np.array(param_values[param_name]) for param_name in param_names]
            # take care of rounding errors and negative zeros
            combinations = [(arr.round(decimals=8) + 0.0) if arr[0] is not None else arr for arr in combinations]
            # take care of duplicate ploints
            combinations = [np.unique(arr) for arr in combinations]
            combinations = itertools.product(*combinations)

            for combination in combinations:
                param_point = {k:v for k,v in zip(param_names, combination)}
                param_points.append(param_point)
        return param_points
    
    @staticmethod
    def val_encode_parameters(parameters:Dict):
        tokens = []
        for param, value in parameters.items():
            if value is None:
                token = f"{param}"
            else:
                token = f"{param}={round(value, 8)}"
            tokens.append(token)
        return ",".join(tokens)
    
    @staticmethod
    def str_encode_parameters(parameters:Dict):
        encoded_str_list = []
        for param, value in parameters.items():
            if isinstance(value, float):
                value = str_encode_value(round(value, 8))
            encoded_str = f"{param}_{value}"
            encoded_str_list.append(encoded_str)
        return "_".join(sorted(encoded_str_list))

    def get_external_param_points(self, dirname:str="",
                                  filter_expr:Optional[str]=None,
                                  exclude_expr:Optional[str]=None):
        if os.path.isdir(dirname):
            fnames = glob.glob(os.path.join(dirname, '*'))
        else:
            fnames = glob.glob(dirname)
        param_points = []
        for fname in fnames:
            basename = os.path.basename(fname)
            match = self.fname_regex.match(basename)
            if not match:
                continue
            point = {'filename': fname, 'parameters':{}}
            point['basename'] = basename.split('.')[0]
            for key, value in match.groupdict().items():
                parser = self.attribute_parser[key]
                point['parameters'][key] = parser(value)
            param_points.append(point)
        attributes = list(self.attribute_parser)
        param_points = self.sort_param_points(param_points, attributes)
        selected_param_points = self.select_param_points(param_points, filter_expr,
                                                         exclude_expr, dict_key="parameters")
        return selected_param_points
    
    def get_internal_param_points(self, filter_expr:Optional[str]=None,
                                  exclude_expr:Optional[str]=None):
        param_points = self.parse_param_str(self.param_str)
        selected_param_points = self.select_param_points(param_points, filter_expr,
                                                         exclude_expr)
        return selected_param_points
    
    @staticmethod
    def parse_filter_expr(expr:str):
        if expr is None:
            return []
        expr = remove_whitespace(expr)
        match_conditions = []
        tokens = expr.split(";")
        for token in tokens:
            match_condition = {}
            param_conditions = token.split(",")
            for param_condition in param_conditions:
                subtokens = param_condition.split("=")
                if len(subtokens) != 2:
                    raise RuntimeError(f"invalid parameter condition `{param_condition}` in the "
                                       f"filter/exclude expression `{expr}`")
                param_name, condition = subtokens
                if "*" in condition:
                    match_condition[param_name] = partial(MatchMode.Wildcard.match_func, y=condition)
                else:
                    match_condition[param_name] = partial(MatchMode.Numerical.match_func, y=condition)
            match_conditions.append(match_condition)
        return match_conditions
    
    @staticmethod
    def select_param_point(parameters, conditions:List[Dict], default:bool=True):
        selected = not default
        for condition in conditions:
            for param_name, match_func in condition.items():
                if param_name not in parameters:
                    continue
                param_value = parameters[param_name]
                if not match_func(param_value):
                    return not default
        return default
    
    @staticmethod
    def select_param_point(parameters, conditions:List[Dict], default:bool=True):
        for condition in conditions:
            for param_name, match_func in condition.items():
                # consider not matched if any of the param-specific condition not satisfied
                if (param_name not in parameters) or (not match_func(parameters[param_name])):
                    break
            else:
                return default
        return not default
    
    def select_param_points(self, param_points:List,
                            filter_expr:Optional[str]=None,
                            exclude_expr:Optional[str]=None,
                            dict_key:Optional=None):
        selected_param_points = []
        filter_conditions = self.parse_filter_expr(filter_expr)
        exclude_conditions = self.parse_filter_expr(exclude_expr)
        for param_point in param_points:
            if dict_key:
                parameters = param_point[dict_key]
            else:
                parameters = param_point               
            keep = True
            if filter_conditions:
                keep = self.select_param_point(parameters, filter_conditions, default=True)
            if exclude_conditions:
                keep &= not (self.select_param_point(parameters, exclude_conditions, default=False))
            if keep:
                selected_param_points.append(param_point)
        return selected_param_points
    
    def sanity_check(self, internal_param_points:Dict, external_param_points:Dict):
        if not self.allow_none:
            for int_point in internal_param_points:
                for param_name, param_value in int_point.items():
                    if param_value is None:
                        raise RuntimeError(f"profiled parameter is not allowed ({param_name})")
    
    def get_param_points(self, dirname:str="",
                         filter_expr:Optional[str]=None,
                         exclude_expr:Optional[str]=None):
        external_param_points = self.get_external_param_points(dirname, filter_expr, exclude_expr)     
        internal_param_points = self.get_internal_param_points(filter_expr, exclude_expr)
        self.sanity_check(internal_param_points, external_param_points)
        if len(internal_param_points) == 0:
            internal_param_points = [{}]
        param_points = []
        for ext_point in external_param_points:
            fname = ext_point['filename']
            basename = ext_point['basename']
            ext_params = ext_point['parameters']
            for int_params in internal_param_points:
                param_point = {}
                param_point['filename'] = fname
                param_point['basename'] = basename
                param_point['external_parameters'] = {**ext_params}
                param_point['internal_parameters'] = {**int_params}
                param_points.append(param_point)
        return param_points


from typing import Optional
import re

from .rooproc_rdf_action import RooProcRDFAction

class RooProcDefine(RooProcRDFAction):
    
    def __init__(self, name:str, expression:str):
        super().__init__(name=name, expression=expression)
        
    @classmethod
    def parse(cls, main_text:str, block_text:Optional[str]=None):
        result = re.search(r"^\s*(\w+)\s*=(.*)", main_text)
        if not result:
            raise RuntimeError(f"invalid expression {main_text}")
        name = result.group(1)
        expression = result.group(2)
        return cls(name=name, expression=expression)        
        
    def _execute(self, rdf:"ROOT.RDataFrame", **params):
        name = params['name']
        expression = params['expression']
        rdf_next = rdf.Define(name, expression)
        return rdf_next